// -*- Mode: c++ -*-
// $Id: Face.cpp,v 1.4 2000/04/28 22:14:20 ps Exp $
// $Source: /ufs/multires/projects//src/meshfilter/Face.cpp,v $

/**
 * Copyright 1995-2000, Peter Schr\"oder, California Institute of
 * Technology (Caltech)
 *
 * April 26, 2000
 * All Rights Reserved
 *
 * The use of this software is permitted for non-commercial, educational,
 * and research use only. The software and/or related materials are
 * provided "as-is" without warranty of any kind including any warranties
 * of performance or merchantability or fitness for a particular use or
 * purpose (as set forth in UCC 23212-2313) or for any purpose
 * whatsoever, for the licensed product, however used. In no event shall
 * Caltech be liable for any damages and/or costs, including but not
 * limited to incidental or consequential damages of any kind, including
 * economic damage or injury to property and lost profits, regardless of
 * whether Caltech shall be advised, have reason to know, or in fact
 * shall know of the possibility. User bears all risk relating to quality
 * and performance of the software and/or related materials.
 *
 * Any use other than non-commercial, educational, or research, or any
 * redistribution in original or modified form requires prior written
 * authorization from the copyright holder.
 */

#include <assert.h>
#include "Face.h"
#include "Vert.h"
#include "V4Vert.h"
#include "IV.h"

// a small helper struct
struct VVPair{
  VVPair(Vert * vt, V4Vert * v4vt) : v(vt), v4(v4vt) {}
  Vert * v;
  V4Vert * v4;
};

Face::Face(Vert * verts[], unsigned int size)
  : m_v()
{
  // the centroid of the Face, which is the shared NW corner of all V4Vert
  m_c = new CVec(0, 0, 0); assert(m_c);
  unsigned int i;
  for( i = 0; i < size; i++ ){
    assert(verts[i]);
    V4Vert * v = new V4Vert(verts[i]); assert(v);
    m_v.push_back(new struct VVPair(verts[i], v));
    verts[i]->AddChild(v);
  }

  // link the N and W of every V4Vert
  for( i = 0; i < size; i++ ){
    assert( m_v[i] && m_v[i]->v4 && m_v[(i+1)%size] && m_v[(i-1+size)%size] );
    m_v[i]->v4->Link(N, m_v[(i+1)%size]->v4); 
    m_v[i]->v4->Link(W, m_v[(i-1+size)%size]->v4); 
    assert(m_v[i]->v4->Neighbor(N) && m_v[i]->v4->Neighbor(W));
    m_v[i]->v4->SetCorner(NW, m_c);
    assert(m_v[i]->v4->Corner(NW));
  }
}

Face::~Face()
{
  std::vector<struct VVPair *>::iterator fb = m_v.begin(), fi = fb, fe = m_v.end();
  for( ; fi != fe; fi++ ){
    // unlink the pointers only. The central corner will be taken care of
    // by the Face. other corners will be taken care of by the Verts
    (*fi)->v = NULL;
    (*fi)->v4 = NULL;
    delete (*fi);
  }
  delete m_c;
}

void
Face::Update( void )
{
  // the central corner is the average of the V4Verts in the Face
  (*m_c) = CVec(0, 0, 0);
  std::vector<struct VVPair *>::const_iterator fb = m_v.begin(), fi = fb, fe = m_v.end();
  for( ; fi != fe; fi++){
    assert( *fi && (*fi)->v4 );
    (*m_c) += (*fi)->v4->Loc();
  }
  (*m_c) /= m_v.size()*1.0;
}

// if this Face is on the one ring of the boundary, return true.
bool
Face::BoundaryOneRing( void ) const
{
  std::vector<struct VVPair *>::const_iterator vb = m_v.begin(), vi = vb, ve = m_v.end();
  for( ; vi != ve; vi++){
    assert( *vi && (*vi)->v );
    if( !(*vi)->v->CompleteOneRing() ) return true;
  }
  return false;
}


void
Face::WriteFace(Scheme s, ostream * os, std::map<const CVec*, unsigned long>* vmap) const
{
  // for DooSabin, all children are DooSabin points
  // for Biquartic, if this Face in on the one ring of the boundary, then
  // some of the children may not be real Biquartic points
  if( s == DS || ( s== BIQ && !BoundaryOneRing() ) ){
    assert( *m_v.begin() && (*m_v.begin())->v4 );
    V4Vert * vb = (*m_v.begin())->v4, *v1 = vb, *v2 = v1->Neighbor(N);
    do{
      assert( v2 && (vmap->find(&v1->Loc()) != vmap->end()) );
      *os << (*vmap)[&v1->Loc()] << " ";
      v1 = v2;
      v2 = v1->Neighbor(N);
    }while(v1 != vb);
    *os << END_FACE_INDEX << endl;
  }
}

// a variation of WriteFace for iv output
void Face::WriteFaceIV(Scheme s, ostream * os, std::map<const CVec*, unsigned long>* vmap, const char * indent) const
{
  if( s == DS || ( s== BIQ && !BoundaryOneRing() ) ){
    assert( *m_v.begin() && (*m_v.begin())->v4 );
    V4Vert * vb = (*m_v.begin())->v4, *v1 = vb, *v2 = v1->Neighbor(N);
    *os << indent;
    do{
      assert( v2 && (vmap->find(&v1->Loc()) != vmap->end()) );
      *os << (*vmap)[&v1->Loc()] << ", ";
      v1 = v2;
      v2 = v1->Neighbor(N);
    }while(v1 != vb);
    *os << END_FACE_INDEX << ", " << endl;
  }
}

