// -*- Mode: c++ -*-
// $Id: Mesh.cpp,v 1.4 2000/04/28 22:14:21 ps Exp $
// $Source: /ufs/multires/projects//src/meshfilter/Mesh.cpp,v $

/**
 * Copyright 1995-2000, Peter Schr\"oder, California Institute of
 * Technology (Caltech)
 *
 * April 26, 2000
 * All Rights Reserved
 *
 * The use of this software is permitted for non-commercial, educational,
 * and research use only. The software and/or related materials are
 * provided "as-is" without warranty of any kind including any warranties
 * of performance or merchantability or fitness for a particular use or
 * purpose (as set forth in UCC 23212-2313) or for any purpose
 * whatsoever, for the licensed product, however used. In no event shall
 * Caltech be liable for any damages and/or costs, including but not
 * limited to incidental or consequential damages of any kind, including
 * economic damage or injury to property and lost profits, regardless of
 * whether Caltech shall be advised, have reason to know, or in fact
 * shall know of the possibility. User bears all risk relating to quality
 * and performance of the software and/or related materials.
 *
 * Any use other than non-commercial, educational, or research, or any
 * redistribution in original or modified form requires prior written
 * authorization from the copyright holder.
 */

#include "Mesh.h"
#include "Vert.h"
#include "Face.h"
#include "IV.h"

Mesh::~Mesh( void )
{
  // delete all Verts and Faces
  unsigned int size = m_vc.size(), i;
  for( i = 0; i < size; i++ ) delete m_vc[i];
  size = m_fc.size();
  for( i = 0; i < size; i++ ) delete m_fc[i];
}

void
Mesh::Subdiv( void )
{
  unsigned int vsize = m_vc.size(), fsize = m_fc.size(), i;
  // averaging once to get DooSabin
  // averaging twice to get Biquartic
  int av = (scheme == DS) ? 1 : ((scheme == BIQ) ? 2 : 0);
  for(int j = 0; j < av; j++){
    for( i = 0; i < vsize; i++) m_vc[i]->Update();
    for( i = 0; i < fsize; i++) m_fc[i]->Update();
    for( i = 0; i < vsize; i++) m_vc[i]->Subdiv();
  }
}

// find the node of the specified type
template <class NodeType>
QvNode *
findNode( QvNode * root, NodeType * dummy ) {
  QvNode * res = 0;
  QvGroup * grp = 0;
  if( (grp = dynamic_cast<QvGroup *>(root)) ){
    int i, num = grp->getNumChildren();
    for( i = 0; i < num; i++ )
      { if( (res = findNode(grp->getChild(i), dummy)) ) return res; }
    return NULL;
  }
  else{
    return dynamic_cast<NodeType*>(root);
  }
}

// read in an iv file
bool
readIVFile( FILE * fp, Mesh& m )
{
  if( !fp ) return false;
  QvDB::init();
  QvInput scene;
  scene.setFilePointer( fp );

  QvSeparator * root = 0;
  QvNode * node = 0;
  QvBool res = QvDB::read( &scene, node );
  if( !res || !node ) return false;
  if( dynamic_cast<QvSeparator *>(node) ){
    root = (QvSeparator *) node;
  }
  else{
    root = new QvSeparator;
    root->getChildren()->append( node );
  }
  assert( root );

  root->ref();

  while( node ){
    node = 0;
    res = QvDB::read( &scene, node );
    if( res && node ) root->getChildren()->append( node );
  }

  fclose( fp );

  QvCoordinate3 * coords = 0;
  QvIndexedFaceSet * faces = 0;

  QvNode * path = findNode( root, (QvCoordinate3*)0 );

  if( !path ) { root->unref(); return false; }
  else coords = (QvCoordinate3*) path;
  path = findNode( root, (QvIndexedFaceSet*)0 );
  if( !path ) { root->unref(); return false; }
  else faces = (QvIndexedFaceSet*) path;
  assert( coords && faces );

  coords->ref(); faces->ref();

  unsigned int vsize, isize, i;
  vsize = coords->point.num;
  m.m_vc.reserve( vsize );
  for( i = 0; i < vsize; ++i ){
    Vert * vt = new Vert( CVec( coords->point.values[3*i+0],
				coords->point.values[3*i+1],
				coords->point.values[3*i+2] ) ); 
    assert( vt );
    m.m_vc.push_back( vt );
  }
  isize = faces->coordIndex.num; i = 0;
  while( i < isize ){
    static Vert * vts[__MAXFACESIZE];
    int j = 0, index;
    while( (index = faces->coordIndex.values[i++]) != END_FACE_INDEX ){
      assert( j < __MAXFACESIZE && i < isize );
      vts[j++] = m.m_vc[index];
    }
    Face *f = new Face( vts, j ); assert( f );
    m.m_fc.push_back( f );
  }

  // Build the connectivity
  std::vector<Vert *>::iterator vb = m.m_vc.begin(), vi, ve = m.m_vc.end();
  for( vi = vb; vi != ve; vi++)
    (*vi)->BuildTopology();

  coords->unref(); faces->unref(); root->unref();

  return true;
}

// output the valence-4 mesh to the ostream
// the format is so simple that it could be extended for the convenience
// of the user
ostream&
Mesh::WritePrimalPolygon( ostream& os ) const
{
  unsigned long vno = 0, pno = 0;
  std::vector<Vert *>::const_iterator vb = m_vc.begin(), vi, ve = m_vc.end();
  std::vector<Face *>::const_iterator fb = m_fc.begin(), fi, fe = m_fc.end();
  for( vi = vb; vi != ve; vi++ ) vno += (*vi)->CountVerts(scheme);

  for( vi = vb; vi != ve; vi++ ) pno += (*vi)->CountPolygons( scheme );
  for( fi = fb; fi != fe; fi++ ) pno += (*fi)->CountPolygons( scheme );

  if( !( os << vno << " " << pno << endl ) ) return os;

  std::map<const CVec*, unsigned long> vmap;
  unsigned long i = 0;
  os.precision(16);
  for( vi = vb; vi != ve; vi++ ) i = (*vi)->WriteVerts(scheme, &os, &vmap, i);

  assert( i == vno );

  for( vi = vb; vi != ve; vi++ ) (*vi)->WriteFace(scheme, &os, &vmap);
  for( fi = fb; fi != fe; fi++ ) (*fi)->WriteFace(scheme, &os, &vmap);
  
  return os;
}

// extends the output for iv format
ostream&
Mesh::WritePrimalPolygonIV( ostream& os ) const
{
  std::vector<Vert *>::const_iterator vb = m_vc.begin(), vi, ve = m_vc.end();
  std::vector<Face *>::const_iterator fb = m_fc.begin(), fi, fe = m_fc.end();

  if( !( os << IV_HEADER << endl << endl ) ) return os;
  if( !( os << "Separator {" << endl ) ) return os;
  if( !( os << INDENT1 << "Coordinate3 {" << endl ) ) return os;
  if( !( os << INDENT2 << "point [" << endl ) ) return os;

  std::map<const CVec*, unsigned long> vmap;
  unsigned long i = 0;
  os.precision(16);
  for( vi = vb; vi != ve; vi++ ) i = (*vi)->WriteVertsIV(scheme, &os, &vmap, i, INDENT3);

  assert( i = vmap.size() );

  if( !( os << INDENT2 << "]" << endl ) ) return os; // point [
  if( !( os << INDENT1 << "}" << endl ) ) return os; // Coordinate3 {
  if( !( os << INDENT1 << "IndexedFaceSet {" << endl ) ) return os;
  if( !( os << INDENT2 << "coordIndex [" << endl ) ) return os;

  for( vi = vb; vi != ve; vi++ ) {
    assert( *vi );
    (*vi)->WriteFaceIV(scheme, &os, &vmap, INDENT3);
  }
  for( fi = fb; fi != fe; fi++ ) (*fi)->WriteFaceIV(scheme, &os, &vmap, INDENT3);
  
  if( !( os << INDENT2 << "]" << endl ) ) return os; // coordIndex [
  if( !( os << INDENT1 << "}" << endl ) ) return os; // IndexedFaceSet {

  os << "}" << endl; // Separator {

  return os;
}

// for smf input
istream& 
operator >>( istream& is, Mesh& m)
{
  unsigned long vsize = 0, fsize = 0;
  if( !(is >> vsize >> fsize) ) return is;
  m.m_vc.reserve(vsize); m.m_fc.reserve(fsize);
  unsigned long i;
  for( i = 0; i < vsize; ++i ){
    static CVec p;
    if( !( is >> p ) ) { return is; }
    Vert * v = new Vert(p); assert(v);
    m.m_vc.push_back(v);
  }

  for( i = 0; i < fsize; ++i ){
    static Vert* vs[__MAXFACESIZE];
    unsigned int j = 0;
    int index;
    while( ( is >> index ) && ( index != END_FACE_INDEX ) ){
      assert( j < __MAXFACESIZE );
      vs[j++] = m.m_vc[index];
    }
    Face * f = new Face(vs, j); assert( f );
    m.m_fc.push_back(f);
  }

  // build the connectivity
  std::vector<Vert *>::iterator vb = m.m_vc.begin(), vi, ve = m.m_vc.end();
  for( vi = vb; vi != ve; vi++)
    (*vi)->BuildTopology();
  return is;
}

