// -*- Mode: c++ -*-
// $Id: Vert.cpp,v 1.6 2000/04/28 22:14:21 ps Exp $
// $Source: /ufs/multires/projects//src/meshfilter/Vert.cpp,v $

/**
 * Copyright 1995-2000, Peter Schr\"oder, California Institute of
 * Technology (Caltech)
 *
 * April 26, 2000
 * All Rights Reserved
 *
 * The use of this software is permitted for non-commercial, educational,
 * and research use only. The software and/or related materials are
 * provided "as-is" without warranty of any kind including any warranties
 * of performance or merchantability or fitness for a particular use or
 * purpose (as set forth in UCC 23212-2313) or for any purpose
 * whatsoever, for the licensed product, however used. In no event shall
 * Caltech be liable for any damages and/or costs, including but not
 * limited to incidental or consequential damages of any kind, including
 * economic damage or injury to property and lost profits, regardless of
 * whether Caltech shall be advised, have reason to know, or in fact
 * shall know of the possibility. User bears all risk relating to quality
 * and performance of the software and/or related materials.
 *
 * Any use other than non-commercial, educational, or research, or any
 * redistribution in original or modified form requires prior written
 * authorization from the copyright holder.
 */

#include "Vert.h"
#include "V4Vert.h"
#include "Face.h"
#include "IV.h"
#include <assert.h>

// a small helper struct
struct V4OPair{
  V4OPair(V4Vert * v, bool ob) : c(v), o(ob) {}
  V4Vert * c;
  bool o;
};

// the array mapping a Direction or a QTreeName to its counterpart
// in a reflected orientation
static Direction DP[2][4] = { {W, N, E, S},  // the reflected orientation
			      {N, W, S, E} };// the "normal" orientation
static QTreeName NP[2][4] = { {NW, NE, SE, SW},  // the reflected orientation
			      {NW, SW, SE, NE} };// the "normal" orientation

void 
Vert::AddChild(V4Vert * c){
  assert(c); m_v.push_back(new struct V4OPair(c, true));
}

Vert::~Vert()
{
  // Vert cleans up everything, including all its children and their corners
  // except the centroid of the Faces
  unsigned int size = m_v.size(), i;
  V4Vert * tv, *v;
  bool to, o;
  assert( m_v[0] );
  v = m_v[0]->c; o = m_v[0]->o; assert( v && v->Corner(NP[o][SE]) );
  // delete the corner that "occupies the same slot" of this Vert
  // this corner is shared by all children
  delete v->Corner(NP[o][SE]);
  for(i = 0; i < size; i++){
    assert( m_v[i] );
    v = m_v[i]->c; o = m_v[i]->o; assert( v );
    // unlink the centroid of the Faces and the shared corner
    v->SetCorner(NP[o][NW], NULL);
    v->SetCorner(NP[o][SE], NULL);
    // the boundary
    if( !v->Neighbor(DP[o][E]) && v->Corner(NP[o][NE]) ){
      tv = v->Neighbor(DP[o][N]);
      assert( tv && tv->Corner(NP[o][SW]) );
      delete v->Corner(NP[o][NE]);
      v->SetCorner(NP[o][NE], NULL);
      tv->SetCorner(NP[o][SW], NULL);
    }
    // the boundary
    if( !v->Neighbor(DP[o][S]) && v->Corner(NP[o][SW]) ){
      tv = v->Neighbor(DP[o][W]);
      assert( tv && tv->Corner(NP[o][NE]) );
      delete v->Corner(NP[o][SW]);
      v->SetCorner(NP[o][SW], NULL);
      tv->SetCorner(NP[o][NE], NULL);
    }
    // the internal edges
    if( (tv = v->Neighbor(DP[o][S])) && v->Corner(NP[o][SW]) ){
      assert( m_v[(i+1)%size] && ( tv == m_v[(i+1)%size]->c ) );
      to = m_v[(i+1)%size]->o;
      assert( tv->Corner(NP[to][NE]) && v->Neighbor(DP[o][W]) && 
	      tv->Neighbor(DP[to][N]) && 
	      v->Neighbor(DP[o][W])->Corner(NP[o][NE]) && 
	      tv->Neighbor(DP[to][N])->Corner(NP[to][SW]) &&
	      v->Neighbor(DP[o][W])->Parent() == tv->Neighbor(DP[to][N])->Parent() );
      delete v->Corner(NP[o][SW]);
      v->SetCorner(NP[o][SW], NULL);
      tv->SetCorner(NP[to][NE], NULL);
      v->Neighbor(DP[o][W])->SetCorner(NP[o][NE], NULL);
      tv->Neighbor(DP[to][N])->SetCorner(NP[to][SW], NULL);
      v->Neighbor(DP[o][W])->UnLink(DP[o][N]);
      tv->UnLink(DP[to][E]);
    }
  }
  // finally delete all children
  for(i = 0; i < size; i++){
    delete m_v[i]->c;
    delete m_v[i];
  }
}

// build the connectivity between children, create corner slots and link them
// up with the children
void 
Vert::BuildTopology()
{
  std::vector<struct V4OPair *> tmp;
  tmp.push_back(m_v.front());
  m_v.erase(m_v.begin());
  unsigned int size = m_v.size(), i;

  assert( tmp.back()->c->Neighbor(W) && tmp.front()->c->Neighbor(N) );
  // first child's orientation is set to be the "normal" orientation
  Vert * tail = tmp.back()->c->Neighbor(W)->Parent();
  Vert * head = tmp.front()->c->Neighbor(N)->Parent();
  assert( tail && head);
  // in each loop step, find the Face which is adjacent to the current sector
  for(i = 0; i < size; i++){
    std::vector<struct V4OPair *>::iterator fb = m_v.begin(), fi = fb, fe=m_v.end();
    for( ; fi != fe; fi++ ){
      assert( (*fi)->c->Neighbor(W) && (*fi)->c->Neighbor(N) );
      Vert * prev = (*fi)->c->Neighbor(W)->Parent();
      Vert * next = (*fi)->c->Neighbor(N)->Parent();
      assert( prev && next );
      if(prev == head){
	tmp.insert(tmp.begin(), *fi); m_v.erase(fi); head = next;
	break;
      }
      else if(next == tail){
	tmp.push_back(*fi); m_v.erase(fi); tail = prev;
	break;
      }
      else if(prev == tail){ // the new child has reflected configuration
	(*fi)->o = false;    // flip the orientation
	tmp.push_back(*fi); m_v.erase(fi); tail = next;
	break;
      }
      else if(next == head){ // the new child has relfectd configuration
	(*fi)->o = false;    // flip the orientation
	tmp.insert(tmp.begin(), *fi); m_v.erase(fi); head = prev;
	break;
      }
    }
  }

  // the Vert is required to have a continuous neighborhood, otherwise this
  // assertion fails.
  assert(m_v.empty() && tmp.size() == size+1);

  // dump the Faces back to the original container
  size = tmp.size();
  std::vector<struct V4OPair *>::const_iterator fb = tmp.begin(), fi = fb, fe = tmp.end();
  for( ; fi != fe; fi++ )
    m_v.push_back(*fi);
  assert(m_v.size() == tmp.size());

  V4Vert * tv, *v;
  bool to, o;
  for(i = 0; i < size; i++){
    v = m_v[i]->c; tv = m_v[(i+1)%size]->c;
    o = m_v[i]->o, to = m_v[(i+1)%size]->o;
    assert(tv && tv->Neighbor(N) && tv->Neighbor(W) && 
	   v && v->Neighbor(W) && v->Neighbor(N));
    // the two children's faces share an edge
    if( tv->Neighbor(DP[to][N])->Parent() == v->Neighbor(DP[o][W])->Parent() ){
      // link children's neighbors
      v->Link( DP[o][S], tv ); assert( v->Neighbor(DP[o][S]) == tv );
      tv->Link( DP[to][E], v ); assert( tv->Neighbor(DP[to][E]) == v );
      if( this > v->Neighbor(DP[o][W])->Parent() ){
	assert( !v->Corner(NP[o][SW]) && !tv->Corner(NP[to][NE]) &&
		!v->Neighbor(DP[o][W])->Corner(NP[o][NE]) && 
		!tv->Neighbor(DP[to][N])->Corner(NP[to][SW]) );
	// allocation a new corner
	CVec * c = new CVec(0, 0, 0); assert(c);
	v->SetCorner(NP[o][SW], c); assert(v->Corner(NP[o][SW]));
	tv->SetCorner(NP[to][NE], c);
	v->Neighbor(DP[o][W])->SetCorner(NP[o][NE], c);
	tv->Neighbor(DP[to][N])->SetCorner(NP[to][SW], c);
	assert( v->Corner(NP[o][SW]) == tv->Corner(NP[to][NE]) && 
		v->Corner(NP[o][SW]) == v->Neighbor(DP[o][W])->Corner(NP[o][NE]) &&
		v->Corner(NP[o][SW]) == tv->Neighbor(DP[to][N])->Corner(NP[to][SW]) );
      }
    }
    // this Vert must be on the boundary.
    // the two children's face are both on the boundary
    else{
      if( this > v->Neighbor(DP[o][W])->Parent() ){
	assert( !v->Corner(NP[o][SW]) && !v->Neighbor(DP[o][W])->Corner(NP[o][NE]) );
	// allocation a new corner on the boundary
	CVec * c = new CVec(0, 0, 0); assert(c);
	v->SetCorner(NP[o][SW], c); assert(v->Corner(NP[o][SW]));
	v->Neighbor(DP[o][W])->SetCorner(NP[o][NE], c);
      }
      if( this > tv->Neighbor(DP[to][N])->Parent() ){
	assert( !tv->Corner(NP[to][NE]) && !tv->Neighbor(DP[to][N])->Corner(NP[to][SW]) );
	// allocation a new corner on the boundary
	CVec * c = new CVec(0, 0, 0); assert(c);
	tv->SetCorner(NP[to][NE], c); assert(tv->Corner(NP[to][NE]));
	tv->Neighbor(DP[to][N])->SetCorner(NP[to][SW], c);
      }
    }
  }
  // allocation a corner shared by all children
  CVec * c = new CVec(0, 0, 0); assert(c);
  fb = m_v.begin();
  fi = fb;
  fe = m_v.end();
  for( ; fi != fe; fi++ )
    (*fi)->c->SetCorner(NP[(*fi)->o][SE], c);
}

// compute the geometric location of the corners of all children
// except the centroid of the Faces
void
Vert::Update( void )
{
  unsigned int size = m_v.size(), i;
  assert(m_v[0]->c->AllCorners());
  // initialize the shared corner
  *m_v[0]->c->Corner(NP[m_v[0]->o][SE]) = CVec(0, 0, 0);
  V4Vert * tv, *v;
  bool to, o;
  for( i = 0; i < size; i++ ){
    v = m_v[i]->c; assert( v && v->AllCorners() );
    o = m_v[i]->o;
    *v->Corner(NP[o][SE]) += v->Loc();

    assert( v->Neighbor(N) && v->Neighbor(W) &&
	    v->Neighbor(N)->AllCorners() && v->Neighbor(W)->AllCorners() );
    // this Vert is on the boundary. the corner on the boundary should be
    // the average of two adjacent Verts
    if( !v->Neighbor(DP[o][E]) && this > v->Neighbor(DP[o][N])->Parent() )
      *v->Corner(NP[o][NE]) = (this->Loc() + v->Neighbor(DP[o][N])->Parent()->Loc())/2.0;
    // this Vert is on the boundary. the corner on the boundary should be
    // the average of two adjacent Verts
    if( !v->Neighbor(DP[o][S]) && this > v->Neighbor(DP[o][W])->Parent() )
      *v->Corner(NP[o][SW]) = (this->Loc() + v->Neighbor(DP[o][W])->Parent()->Loc())/2.0;
    if( (tv = v->Neighbor(DP[o][S])) && this > v->Neighbor(DP[o][W])->Parent() ){
      to = m_v[(i+1)%size]->o;
      assert( tv->AllCorners() && tv->Neighbor(DP[to][N]) && 
	      tv->Neighbor(DP[to][N])->AllCorners() &&
	      tv->Neighbor(DP[to][N])->Parent() == v->Neighbor(DP[o][W])->Parent() );
      // the corner should be the average of the four incident V4Vert
      *v->Corner(NP[o][SW])=(v->Loc() + v->Neighbor(DP[o][W])->Loc() +
			     tv->Loc() + tv->Neighbor(DP[to][N])->Loc())/4.0;
    }
  }
  // the shared corner should be the average of all children
  *v->Corner(NP[o][SE]) /= size;
}

// compute the geometric position of the children
void
Vert::Subdiv( void )
{
  std::vector<struct V4OPair *>::iterator fb = m_v.begin(), fi = fb, fe = m_v.end();
  for( ; fi != fe; fi++)
    (*fi)->c->Update();
}

// if this Vert in on the boundary, return false.
bool 
Vert::CompleteOneRing( void ) const
{ 
  assert( m_v.front() );
  V4Vert *f = m_v.front()->c; assert( f );
  bool fo = m_v.front()->o;
  assert( f->Neighbor(DP[fo][N]) );

  assert( m_v.back() );
  V4Vert *b = m_v.back()->c; assert( b );
  bool bo = m_v.back()->o;
  assert( b->Neighbor(DP[bo][W]) );

  return f->Neighbor(DP[fo][N])->Parent() == b->Neighbor(DP[bo][W])->Parent(); 
}

unsigned int
Vert::CountVerts( Scheme scheme ) const
{
  // for DooSabin, all computed children are legal DooSabin points
  // for Biquartic, boundary Verts do not have legal BiQuartic children
  return ( scheme == DS || ( scheme == BIQ && CompleteOneRing() ) ) ? m_v.size() : 0;
}

// count the quadrilaterals between neighboring Verts
unsigned int
Vert::CountPolygons( Scheme scheme ) const
{
  unsigned int cnt = 0, size = m_v.size(), i;
  V4Vert *v, *tv;
  bool o, to;
  if( scheme == DS ){
    if(CompleteOneRing()){ 
      ++cnt; // the central guy
      for(i = 0; i < size; i++){
	assert( m_v[i] );
	v = m_v[i]->c; assert( v );
	o = m_v[i]->o;
	tv = v->Neighbor(DP[o][W]); assert( tv );
	if( this > tv->Parent() ) ++cnt; // the polygon on the edge
      }
    }else{
      for( i = 0; i < size; i++ ){
	assert( m_v[i] && m_v[(i+1)%size] );
	v = m_v[i]->c; tv = m_v[(i+1)%size]->c; 
	o = m_v[i]->o; to = m_v[(i+1)%size]->o;
	assert( tv && tv->Neighbor(DP[to][N]) && v && v->Neighbor(DP[o][W]) );
	if( (tv->Neighbor(DP[to][N])->Parent() 
	     == v->Neighbor(DP[o][W])->Parent()) 
	    && (this > v->Neighbor(DP[o][W])->Parent()) )
	  ++cnt; // the polygon on the edge in the interior
      }
    }
  }
  // BIQ must be complete-one-ring to have legal Biq points, hence polygons
  else if( scheme == BIQ && CompleteOneRing()){
    ++cnt; // the central guy
    for( i = 0; i < size; i++ ){
      Vert * n;
      assert( m_v[i] );
      v = m_v[i]->c; o = m_v[i]->o;
      assert( v && v->Neighbor(DP[o][W]) && v->Neighbor(DP[o][W])->Parent() );
      if( (n = v->Neighbor(DP[o][W])->Parent())->CompleteOneRing() &&
	  this > n )
	++cnt; // the polygon on the edge
    }
  }

  return cnt;
}

// output children in for smf format
unsigned long
Vert::WriteVerts(Scheme s, ostream * os, std::map<const CVec*, unsigned long>* vmap, unsigned long seq) const
{
  if( s == DS || (s == BIQ && CompleteOneRing())){
    std::vector<struct V4OPair *>::const_iterator vb = m_v.begin(), vi, ve = m_v.end();
    for( vi = vb; vi != ve; vi++){
      *os << (*vi)->c->Loc() << endl;
      assert( vmap->find(&(*vi)->c->Loc()) == vmap->end());
      (*vmap)[&(*vi)->c->Loc()] = seq++;
    }
  }
  return seq;
}

// a variation of WriteVerts for iv output
unsigned long
Vert::WriteVertsIV(Scheme s, ostream * os, std::map<const CVec*, unsigned long>* vmap, unsigned long seq, const char * indent) const
{
  if( s == DS || (s == BIQ && CompleteOneRing())){
    std::vector<struct V4OPair *>::const_iterator vb = m_v.begin(), vi, ve = m_v.end();
    for( vi = vb; vi != ve; vi++){
      assert( (*vi) && (*vi)->c );
      *os << indent << (*vi)->c->Loc() << " ,"<< endl;
      assert( vmap->find(&(*vi)->c->Loc()) == vmap->end());
      (*vmap)[&(*vi)->c->Loc()] = seq++;
    }
  }
  return seq;
}

static void WriteQuad(const V4Vert * v1, const V4Vert * v2, const V4Vert * v3,
		      const V4Vert* v4, ostream * os, 
		      std::map<const CVec *, unsigned long> * vmap)
{
  assert( vmap->find(&v1->Loc()) != vmap->end() );
  assert( vmap->find(&v2->Loc()) != vmap->end() );
  assert( vmap->find(&v3->Loc()) != vmap->end() );
  assert( vmap->find(&v4->Loc()) != vmap->end() );
  *os << (*vmap)[&v1->Loc()] << " " << (*vmap)[&v2->Loc()] << " "
      << (*vmap)[&v3->Loc()] << " " << (*vmap)[&v4->Loc()] << " "
      << END_FACE_INDEX << endl;
}

static void WriteQuadIV(const V4Vert * v1, const V4Vert * v2, 
			const V4Vert * v3, const V4Vert* v4, ostream * os, 
			std::map<const CVec *, unsigned long> * vmap,
			const char * indent)
{
  assert( vmap->find(&v1->Loc()) != vmap->end() );
  assert( vmap->find(&v2->Loc()) != vmap->end() );
  assert( vmap->find(&v3->Loc()) != vmap->end() );
  assert( vmap->find(&v4->Loc()) != vmap->end() );
  *os << indent;
  *os << (*vmap)[&v1->Loc()] << ", " << (*vmap)[&v2->Loc()] << ", "
      << (*vmap)[&v3->Loc()] << ", " << (*vmap)[&v4->Loc()] << ", "
      << END_FACE_INDEX << ", " << endl;
}

// output polygons in smf format
void
Vert::WriteFace(Scheme s, ostream * os, std::map<const CVec*, unsigned long>* vmap) const
{
  unsigned int size = m_v.size(), i;
  V4Vert * v, *tv;
  bool o, to;
  if( s == DS ){
    for(i = 0; i < size; i++){
      assert(m_v[i] && m_v[(i+1)%size]);
      v = m_v[i]->c; tv = m_v[(i+1)%size]->c;
      o = m_v[i]->o; to = m_v[(i+1)%size]->o;
      assert( v && v->Neighbor(DP[o][W]) && tv && tv->Neighbor(DP[to][N]) );
      if( v->Neighbor(DP[o][W])->Parent() == tv->Neighbor(DP[to][N])->Parent()
	  && this > v->Neighbor(DP[o][W])->Parent() )
	// the quad on the edge
	WriteQuad(v, v->Neighbor(DP[o][W]), tv->Neighbor(DP[to][N]), tv, os, vmap);
    }
    // the central guy
    if(CompleteOneRing()){
      for(i = 0; i < size; i++){
	assert(m_v[i] && m_v[i]->c && vmap->find(&m_v[i]->c->Loc()) != vmap->end());
	*os << (*vmap)[&m_v[i]->c->Loc()] << " ";
      }
      *os << END_FACE_INDEX << endl;
    }
  }
  else if( s == BIQ && CompleteOneRing() ) {
    // the central guy
    for(i = 0; i < size; i++){
      assert(m_v[i] && m_v[i]->c && vmap->find(&m_v[i]->c->Loc()) != vmap->end());
      *os << (*vmap)[&m_v[i]->c->Loc()] << " ";
    }
    *os << END_FACE_INDEX << endl;
    for(i = 0; i < size; i++){
      assert(m_v[i] && m_v[(i+1)%size]);
      v = m_v[i]->c; tv = m_v[(i+1)%size]->c;
      o = m_v[i]->o; to = m_v[(i+1)%size]->o;
      assert( v && v->Neighbor(DP[o][W]) && tv && tv->Neighbor(DP[to][N]) &&
	      v->Neighbor(DP[o][W])->Parent() == tv->Neighbor(DP[to][N])->Parent());
      Vert * n;
      if((n = v->Neighbor(DP[o][W])->Parent())->CompleteOneRing() && this > n)
	// the quad on the edge
	WriteQuad(v, v->Neighbor(DP[o][W]), tv->Neighbor(DP[to][N]), tv, os, vmap);
    }
  }
}

// iv output
void
Vert::WriteFaceIV(Scheme s, ostream * os, std::map<const CVec*, unsigned long>* vmap, const char * indent) const
{
  unsigned int size = m_v.size(), i;
  V4Vert * v, *tv;
  bool o, to;
  if( s == DS ){
    for(i = 0; i < size; i++){
      assert(m_v[i] && m_v[(i+1)%size]);
      v = m_v[i]->c; tv = m_v[(i+1)%size]->c;
      o = m_v[i]->o; to = m_v[(i+1)%size]->o;
      assert( v && v->Neighbor(DP[o][W]) && tv && tv->Neighbor(DP[to][N]) );
      if( v->Neighbor(DP[o][W])->Parent() == tv->Neighbor(DP[to][N])->Parent()
	  && this > v->Neighbor(DP[o][W])->Parent() )
	WriteQuadIV(v, v->Neighbor(DP[o][W]), tv->Neighbor(DP[to][N]), tv, os, vmap, indent);
    }
    // the central guy
    if(CompleteOneRing()){
      *os << indent;
      for(i = 0; i < size; i++){
	assert(m_v[i] && m_v[i]->c && vmap->find(&m_v[i]->c->Loc()) != vmap->end());
	*os << (*vmap)[&m_v[i]->c->Loc()] << ", ";
      }
      *os << END_FACE_INDEX << ", " << endl;
    }
  }
  else if( s == BIQ && CompleteOneRing() ) {
    // the central guy
    *os << indent;
    for(i = 0; i < size; i++){
      assert(m_v[i] && m_v[i]->c && vmap->find(&m_v[i]->c->Loc()) != vmap->end());
      *os << (*vmap)[&m_v[i]->c->Loc()] << ", ";
    }
    *os << END_FACE_INDEX << ", " << endl;
    for(i = 0; i < size; i++){
      assert(m_v[i] && m_v[(i+1)%size]);
      v = m_v[i]->c; tv = m_v[(i+1)%size]->c;
      o = m_v[i]->o; to = m_v[(i+1)%size]->o;
      assert( v && v->Neighbor(DP[o][W]) && tv && tv->Neighbor(DP[to][N]) &&
	      v->Neighbor(DP[o][W])->Parent() == tv->Neighbor(DP[to][N])->Parent());
      Vert * n;
      if((n = v->Neighbor(DP[o][W])->Parent())->CompleteOneRing() && this > n)
	WriteQuadIV(v, v->Neighbor(DP[o][W]), tv->Neighbor(DP[to][N]), tv, os, vmap, indent);
    }
  }
}

