// -*- Mode: c++ -*-
// $Id: CFace.cpp,v 1.5 2000/04/28 22:14:43 ps Exp $
// $Source: /ufs/multires/projects//src/CourseQuad/CFace.cpp,v $

/**
 * Copyright 1995-2000, Peter Schr\"oder, California Institute of
 * Technology (Caltech) 
 *
 * April 26, 2000
 * All Rights Reserved
 * 
 * The use of this software is permitted for non-commercial, educational,
 * and research use only. The software and/or related materials are
 * provided "as-is" without warranty of any kind including any warranties
 * of performance or merchantability or fitness for a particular use or
 * purpose (as set forth in UCC 23212-2313) or for any purpose
 * whatsoever, for the licensed product, however used. In no event shall
 * Caltech be liable for any damages and/or costs, including but not
 * limited to incidental or consequential damages of any kind, including
 * economic damage or injury to property and lost profits, regardless of
 * whether Caltech shall be advised, have reason to know, or in fact
 * shall know of the possibility. User bears all risk relating to quality
 * and performance of the software and/or related materials.
 * 
 * Any use other than non-commercial, educational, or research, or any
 * redistribution in original or modified form requires prior written
 * authorization from the copyright holder.
 */

#include <algorithm>
#include "CFace.h"
#include "cvec.h"
#include "Cp.h"
#include "QTree.h"

CFace::CFace(Cp * cps[], unsigned int sides)
  : m_cpc(sides)
{
  assert(sides > 0);
  // transfer control points
  std::copy( cps, cps+sides, m_cpc.begin() );
  // compute centroid for later use
  CVec c(0,0,0);
  for( unsigned int i = 0; i < sides; ++i ) c += cps[i]->Root()->loc();
  cent = new CVec( c/sides ); assert( cent );
}

CFace::~CFace( void )
{
  delete cent;
}

/*
 * link the current Cp to its neighbors
 */
void
CFace::BuildTopology( void )
{
  unsigned int size = m_cpc.size();
  for( unsigned int i = 0; i < size; ++i ){
    Cp * p = m_cpc[(i+size-1)%size];
    Cp * c = m_cpc[i];
    Cp * n = m_cpc[(i+1)%size];

    switch( ( !!c->Root()->Corner(NW) << 3 ) |
	    ( !!c->Root()->Corner(SW) << 2 ) |
	    ( !!c->Root()->Corner(SE) << 1 ) |
	    ( !!c->Root()->Corner(NE) << 0 ) ){
    case 0:  // no connection
      assert(!c->Neighbor(N) && !c->Neighbor(W) && !c->Neighbor(S) && !c->Neighbor(E) );
      // always link NW first
      c->Link(n, N); c->Link(p, W); c->Root()->SetCorner(NW, cent);
      break;
    case 8:  // NW
      assert(c->Neighbor(N) && c->Neighbor(W) && !c->Neighbor(S) && !c->Neighbor(E) );
      
      if     ( n == c->Neighbor(W) ) { c->Link(p, S); c->Root()->SetCorner(SW, cent); } 
      else if( p == c->Neighbor(W) ) { c->Link(n, S); c->Root()->SetCorner(SW, cent); }
      else if( n == c->Neighbor(N) ) { c->Link(p, E); c->Root()->SetCorner(NE, cent); }
      else if( p == c->Neighbor(N) ) { c->Link(n, E); c->Root()->SetCorner(NE, cent); }
      else            { c->Link(p, E); c->Link(n, S); c->Root()->SetCorner(SE, cent); }
      
      break;
    case 12: // NW, SW
      assert(c->Neighbor(N) && c->Neighbor(W) && c->Neighbor(S) && !c->Neighbor(E) );
      
      if     ( n == c->Neighbor(S) ) { c->Link(p, E); c->Root()->SetCorner(SE, cent); }
      else if( p == c->Neighbor(S) ) { c->Link(n, E); c->Root()->SetCorner(SE, cent); }
      else if( n == c->Neighbor(N) ) { c->Link(p, E); c->Root()->SetCorner(NE, cent); }
      else { assert( p == c->Neighbor(N) ); c->Link(n, E); c->Root()->SetCorner(NE, cent); }
      
      break;
    case 9:  // NW, NE
      assert(c->Neighbor(N) && c->Neighbor(W) && !c->Neighbor(S) && c->Neighbor(E) );
      
      if     ( n == c->Neighbor(W) ) { c->Link(p, S); c->Root()->SetCorner(SW, cent); }
      else if( p == c->Neighbor(W) ) { c->Link(n, S); c->Root()->SetCorner(SW, cent); }
      else if( n == c->Neighbor(E) ) { c->Link(p, S); c->Root()->SetCorner(SE, cent); }
      else { assert( p == c->Neighbor(E) ); c->Link(n, S); c->Root()->SetCorner(SE, cent); }
      
      break;
    case 10: // NW, SE
      assert(c->Neighbor(N) && c->Neighbor(W) && c->Neighbor(S) && c->Neighbor(E) );
      
      if( p == c->Neighbor(N) ){
	if( n == c->Neighbor(S) ){
	  // need to flip the pious S, E links
	  Cp * t = c->Neighbor(E); c->Link(c->Neighbor(S), E); c->Link(t, S);
	}else assert( n == c->Neighbor(E) );
	c->Root()->SetCorner(NE, cent);
      }else if( p == c->Neighbor(W) ){
	if( n == c->Neighbor(E) ){
	  // need to flip the pious S, E links
	  Cp * t = c->Neighbor(E); c->Link(c->Neighbor(S), E); c->Link(t, S);
	}else assert( n == c->Neighbor(S) );
	c->Root()->SetCorner(SW, cent);
      }else if( p == c->Neighbor(S) ){
	if( n == c->Neighbor(N) ){
	  // need to flip the pious S, E links
	  Cp * t = c->Neighbor(S); c->Link(c->Neighbor(E), S); c->Link(t, E); c->Root()->SetCorner(NE, cent);
	}else{ assert( n == c->Neighbor(W) ); c->Root()->SetCorner(SW, cent); }
      }else{
	assert( p == c->Neighbor(E) );
	if( n == c->Neighbor(W) ){
	  // need to flip the pious S, E links
	  Cp * t = c->Neighbor(E); c->Link(c->Neighbor(S), E); c->Link(t, S); c->Root()->SetCorner(SW, cent);
	}else{ assert( n == c->Neighbor(N) ); c->Root()->SetCorner(NE, cent); }
      }
      
      break;
    case 13: // NW, SW, NE
      assert(c->Neighbor(N) && c->Neighbor(W) && c->Neighbor(S) && c->Neighbor(E) );
      assert( (p == c->Neighbor(S) && n == c->Neighbor(E)) ||
	      (p == c->Neighbor(E) && n == c->Neighbor(S)));
      c->Root()->SetCorner(SE, cent);
      break;
    case 14: // NW, SW, SE
      assert(c->Neighbor(N) && c->Neighbor(W) && c->Neighbor(S) && c->Neighbor(E) );
      assert( (p == c->Neighbor(N) && n == c->Neighbor(E)) || (p == c->Neighbor(E) && n == c->Neighbor(N)));
      c->Root()->SetCorner(NE, cent);
      break;
    case 11: // NW, SE, NE
      assert(c->Neighbor(N) && c->Neighbor(W) && c->Neighbor(S) && c->Neighbor(E) );
      assert( (p == c->Neighbor(S) && n == c->Neighbor(W)) || (p == c->Neighbor(W) && n == c->Neighbor(S)));
      c->Root()->SetCorner(SW, cent);
      break;
    default: die(); // no other configuration
    }
  }
}

/**
 * this method must be called after "BuildTopology" has been called
 * flip the parity bit of orientation when necessary
 */
void
CFace::ConfigLinks( void )
{
  const unsigned int size = m_cpc.size();

  for( unsigned int i = 0; i < size; ++i ){
    Cp * p = m_cpc[(i+size-1)%size];
    Cp * c = m_cpc[i];
    Cp * n = m_cpc[(i+1)%size];
    Cp *nn = m_cpc[(i+2)%size];

    const Direction p_c = c->Neighbor(p);
    const Direction n_c = c->Neighbor(n);
    const Direction c_n = n->Neighbor(c);
    const Direction nn_n = n->Neighbor(nn);

    assert( p_c != Oppos(n_c) && c_n != Oppos(nn_n) );

    const bool b1 = (p_c == Next(n_c));
    const bool b2 = (c_n == Next(nn_n));
    // flip the orientation
    if( b1 != b2 ) { c->Flip(n_c); n->Flip(c_n); }
  }
}
