// -*- Mode: c++ -*-
// $Id: CMesh.cpp,v 1.15 2000/04/28 22:14:43 ps Exp $
// $Source: /ufs/multires/projects//src/CourseQuad/CMesh.cpp,v $

/**
 * Copyright 1995-2000, Peter Schr\"oder, California Institute of
 * Technology (Caltech) 
 *
 * April 26, 2000
 * All Rights Reserved
 * 
 * The use of this software is permitted for non-commercial, educational,
 * and research use only. The software and/or related materials are
 * provided "as-is" without warranty of any kind including any warranties
 * of performance or merchantability or fitness for a particular use or
 * purpose (as set forth in UCC 23212-2313) or for any purpose
 * whatsoever, for the licensed product, however used. In no event shall
 * Caltech be liable for any damages and/or costs, including but not
 * limited to incidental or consequential damages of any kind, including
 * economic damage or injury to property and lost profits, regardless of
 * whether Caltech shall be advised, have reason to know, or in fact
 * shall know of the possibility. User bears all risk relating to quality
 * and performance of the software and/or related materials.
 * 
 * Any use other than non-commercial, educational, or research, or any
 * redistribution in original or modified form requires prior written
 * authorization from the copyright holder.
 */

#include <map>
#include <vector>
#include <assert.h>

#include "CMesh.h"
#include "Cp.h"
#include "CFace.h"
#include "QTree.h"
#include "Perms.h"

#include "QvDB.h"
#include "QvNode.h"
#include "QvInput.h"
#include "QvGroup.h"
#include "QvSeparator.h"
#include "QvChildList.h"
#include "QvCoordinate3.h"
#include "QvIndexedFaceSet.h"

#define END_FACE_INDEX -1
#define IV_HEADER "#Inventor V2.1 ascii"
#define INDENT1 "    "
#define INDENT2 "        "
#define INDENT3 "            "

// find the node of the specified type
template <class NodeType>
QvNode *
findNode( QvNode * root, NodeType * dummy ) {
  QvNode * res = 0;
  QvGroup * grp = 0;
  if( (grp = dynamic_cast<QvGroup *>(root)) ){
    int i, num = grp->getNumChildren();
    for( i = 0; i < num; i++ )
      { if( (res = findNode(grp->getChild(i), dummy)) ) return res; }
    return NULL;
  }
  else{
    return dynamic_cast<NodeType*>(root);
  }
}

// read in an iv file
bool
readIVFile( FILE * fp, CMesh& m )
{
  if( !fp ) return false;
  QvDB::init();
  QvInput scene;
  scene.setFilePointer( fp );

  QvSeparator * root = 0;
  QvNode * node = 0;
  QvBool res = QvDB::read( &scene, node );
  if( !res || !node ) return false;
  if( dynamic_cast<QvSeparator *>(node) ){
    root = (QvSeparator *) node;
  }
  else{
    root = new QvSeparator;
    root->getChildren()->append( node );
  }
  assert( root );

  root->ref();

  while( node ){
    node = 0;
    res = QvDB::read( &scene, node );
    if( res && node ) root->getChildren()->append( node );
  }

  fclose( fp );

  QvCoordinate3 * coords = 0;
  QvIndexedFaceSet * faces = 0;

  QvNode * path = findNode( root, (QvCoordinate3*)0 );

  if( !path ) { root->unref(); return false; }
  else coords = (QvCoordinate3*) path;
  path = findNode( root, (QvIndexedFaceSet*)0 );
  if( !path ) { root->unref(); return false; }
  else faces = (QvIndexedFaceSet*) path;
  assert( coords && faces );

  coords->ref(); faces->ref();

  unsigned int vsize, isize, i;
  vsize = coords->point.num;
  m.m_cpc.reserve( vsize );
  for( i = 0; i < vsize; ++i ){
    Cp * cp = new Cp( CVec( coords->point.values[3*i+0],
                            coords->point.values[3*i+1],
                            coords->point.values[3*i+2] ),
                      m.m_s ); assert( cp );
    m.m_cpc.push_back( cp );
  }
  isize = faces->coordIndex.num; i = 0;
  while( i < isize ){
    static Cp * cps[__MAXFACESIZE];
    int j = 0, index;
    while( (index = faces->coordIndex.values[i++]) != END_FACE_INDEX ){
      assert( j < __MAXFACESIZE && i < isize );
      cps[j++] = m.m_cpc[index];
    }
    CFace *f = new CFace( cps, j ); assert( f );
    m.m_fc.push_back( f );
  }

  // BuildTopology
  unsigned int size = m.m_fc.size();
  // the order of the two function calls is important
  for( i = 0; i < size; ++i ) m.m_fc[i]->BuildTopology();
  for( i = 0; i < size; ++i ) m.m_fc[i]->ConfigLinks();

  coords->unref(); faces->unref(); root->unref();

  return true;
}

// this is a much simpler reader that does not depend on qvlib. You can
// figure out from the statements below what the expected input format
// is. It should be easy adopt this to any other input format of your
// liking.
istream&
operator>>( istream& is, CMesh& m )
{
  unsigned vsize = 0, psize = 0;
  if( !( is >> vsize >> psize ) ) return is;
  
  // reserve space to store things in.
  m.m_cpc.reserve( vsize ); m.m_fc.reserve( psize );
  unsigned int i;
  for( i = 0; i < vsize; ++i ){
    static CVec p;
    if( !( is >> p ) ){ return is; }
    Cp *cp = new Cp( p, m.m_s ); assert( cp );
    m.m_cpc.push_back( cp );
  }
  for( i = 0; i < psize; ++i ){
    static Cp *cps[__MAXFACESIZE];
    int j = 0, index;
    while( ( is >> index ) && ( index != END_FACE_INDEX ) ){
      assert( j < __MAXFACESIZE );
      cps[j++] = m.m_cpc[index];
    }
    CFace *f = new CFace( cps, j ); assert( f );
    m.m_fc.push_back( f );
  }

  // BuildTopology
  unsigned int size = m.m_fc.size();
  // the order of the two function calls is important
  for( i = 0; i < size; ++i ) m.m_fc[i]->BuildTopology();
  for( i = 0; i < size; ++i ) m.m_fc[i]->ConfigLinks();
  return is;
}

CMesh::~CMesh()
{
  unsigned int size = m_cpc.size(), i;
  for( i = 0; i < size; ++i ) delete m_cpc[i];
  size = m_fc.size();
  for( i = 0; i < size; ++i ) delete m_fc[i];
}

void
CMesh::Subdiv(const unsigned int todepth, const real eps)
{
  m_adaptive = eps > 0;
  if(todepth > m_depth){
    int size = m_cpc.size();
    // subdivide on each control point
    for( int i =0; i < size; i++ ) m_cpc[i]->Root()->Expand(todepth, eps);
    m_depth = todepth;
  }
}

// count the number of vertices as well as initialize them to (0, 0, 0)
struct CountVertices : public unary_function<CVec*, void> {
  CountVertices( void ) : cnt( 0 ) {}
  void operator()( CVec* v ) { assert(v); *v = CVec(0,0,0); ++cnt; }
  unsigned long cnt;
};

// count the number of triangles as well as computing their corners
struct CountTriangles : public unary_function<QTree*, void>{
  CountTriangles( void ) : cnt( 0 ) {}
  void operator()( QTree *t ) {
    if( t->Leaf() ){
      const CVec &c = t->loc()/4;
      if( t->Corner(NW) ) *t->Corner(NW) += c;
      if( t->Corner(SW) ) *t->Corner(SW) += c;
      if( t->Corner(SE) ) *t->Corner(SE) += c;
      if( t->Corner(NE) ) *t->Corner(NE) += c;

      Perms np; int config = 0; const QTree *n;
      // deal with no-neighbor case
      if( ( n = t->Neighbor( W, &np ) ) && ( n->c(np[NE]) && n->c(np[SE]) ) ){
	*n->c(np[NE])->Corner(np[SE]) += c; 
	*n->c(np[SE])->Corner(np[NE]) += c;
	config |= 1;
      }
      if( ( n = t->Neighbor( S, &np ) ) && ( n->c(np[NW]) && n->c(np[NE]) ) ){
	*n->c(np[NW])->Corner(np[NE]) += c;
	*n->c(np[NE])->Corner(np[NW]) += c;
	config |= 2;
      }
      if( ( n = t->Neighbor( E, &np ) ) && ( n->c(np[NW]) && n->c(np[SW]) ) ){
	*n->c(np[NW])->Corner(np[SW]) += c;
	*n->c(np[SW])->Corner(np[NW]) += c;
	config |= 4;
      }
      if( ( n = t->Neighbor( N, &np ) ) && ( n->c(np[SW]) && n->c(np[SE]) ) ){
	*n->c(np[SW])->Corner(np[SE]) += c;
	*n->c(np[SE])->Corner(np[SW]) += c;
	config |= 8;
      }
      
      if( t->Corner(NW) && t->Corner(SW) && t->Corner(SE) && t->Corner(NE) ){
	switch( config ){
	case 0: cnt += 2; break;
	case 1: case 2: case 4: case 8: cnt += 3; break;
	case 3: case 6: case 12: case 9: cnt += 4; break;
	case 5: case 10: cnt += 4; break;
	case 7: case 14: case 13: case 11: cnt += 5; break;
	case 15: cnt += 6; break;
	default: die(); // hellooo?
	}
      }
    }
  }
  unsigned long cnt;
};

struct CountLeafCPs : public unary_function<QTree*, void>{
  CountLeafCPs( void ) : cnt( 0 ) {}
  void operator()( const QTree *t ) { if( t->Leaf() ) ++cnt; }
  unsigned long cnt;
};

// write in the "smf" format
class WriteLeafCPs : public unary_function<QTree*, void>{
public:
  WriteLeafCPs( ostream* os, std::map<const CVec*, unsigned int>* vmap, int seq ) : out( os ), map( vmap ), i( seq ) {}
  virtual void operator() ( QTree *t ) {
    if( t->Leaf() ){
      *out << t->loc() << endl;
      assert( map->find(&t->loc()) == map->end() );
      (*map)[&t->loc()] = i++;
    }
  }
  ostream* out;
  std::map<const CVec*, unsigned int>* map;
  int i;
};

// write in the "iv" format
class WriteLeafCPsIV : public WriteLeafCPs{
public:
  WriteLeafCPsIV( ostream* os, std::map<const CVec*, unsigned int>* vmap, int seq, const char * space ) : WriteLeafCPs( os, vmap, seq ), indent( space ) {}
  virtual void operator()( QTree *t ) {
    if( t->Leaf() ){
      *out << indent << t->loc() << "," << endl;
      assert( map->find(&t->loc()) == map->end() );
      (*map)[&t->loc()] = i++;
    }
  }
  const char * indent;
};

// count the number of quadrilaterals under one Cp or between neighboring Cps
// (other than those shared by more than 3 Cps)
struct CountQuads : public unary_function<QTree*, void>{
  CountQuads( void ) : cnt(0) {}
  // without consideration on boundaries
  void operator()( const QTree * t) {
    if( t->CompleteSetOfChildren() ){
      if ( !t->nw()->CompleteSetOfChildren() ||
	   !t->sw()->CompleteSetOfChildren() ||
	   !t->se()->CompleteSetOfChildren() ||
	   !t->ne()->CompleteSetOfChildren() ){
	++cnt; // center guy
	const Cp * tcp = t->CP();
	const Cp * tncp = t->CP();
	const Cp * twcp = t->CP();
	const Cp * tscp = t->CP();
	const Cp * tecp = t->CP();
	switch( t->n() ){
	case NW:
	  if( t->Root() ){
	    if( tcp >= tecp || !t->Neighbor(E)->CompleteSetOfChildren() ) ++cnt; // E 
	    if( tcp >= tscp || !t->Neighbor(S)->CompleteSetOfChildren() ) ++cnt; // S
	    if( tcp >= tncp || !t->Neighbor(N)->CompleteSetOfChildren() ) ++cnt; // N 
	    if( tcp >= twcp || !t->Neighbor(W)->CompleteSetOfChildren() ) ++cnt; // W 
	  } 
	  else{ 
	    if( tcp >= tncp || !t->Neighbor(N)->CompleteSetOfChildren() ) cnt += 2; // N and NE 
	    if( tcp >= twcp || !t->Neighbor(W)->CompleteSetOfChildren() ) cnt += 2; // W and SW 
	    if( ( tcp == tncp && ( tcp >= twcp || !t->Neighbor(W)->CompleteSetOfChildren())) 
		|| ( tcp == twcp && ( tcp >= tncp || !t->Neighbor(N)->CompleteSetOfChildren()))){ 
	      ++cnt; // NW 
	    } 
	    cnt += 3; // E, S and SE 
	  } 
	  break;
	case SW:
	  // Note: by default, "Root" are NW, so no test on "Root()" anymore
	  ++cnt; // E;
	  if( tcp >= twcp || !t->Neighbor(W)->CompleteSetOfChildren() ) ++cnt; // W
	  if( tcp > tscp || !t->Neighbor(S)->CompleteSetOfChildren() ){
	    cnt += 2; // S and SE
	    if( tcp == twcp ) ++cnt; // SW
	  } 
	  break; 
	case SE:
	  // Note: by default, "Root" are NW, so no test on "Root()" anymore
	  ++cnt; // N
	  if( tcp >= tscp || !t->Neighbor(S)->CompleteSetOfChildren() ) ++cnt; // S
	  if( tcp > tecp || !t->Neighbor(E)->CompleteSetOfChildren() ){
	    cnt += 2; // E and NE
	    if( tcp == tscp ) ++cnt; // SE 
	  } 
	  break; 
	case NE: 
	  // Note: by default, "Root" are NW, so no test on "Root()" anymore
	  if( tcp > tncp || !t->Neighbor(N)->CompleteSetOfChildren() ) ++cnt; // N 
	  if( tcp > tecp || !t->Neighbor(E)->CompleteSetOfChildren() ) ++cnt; // E 
	  break; 
	default: die(); 
	}
      }
    }
    else{
      // deal with the special case on a boundary non-extreme vertex
      // if some of its neighbors are boundary non-extreme vertices
      const Cp * tcp = t->CP();
      if( t->Neighbor(N) && !t->Neighbor(N)->CompleteSetOfChildren() &&
	  t->nw() && t->ne() && ( t->nw()->Leaf() || t->ne()->Leaf() ) &&
	  tcp >= t->Neighbor(N)->CP() ){
	assert( t->Neighbor(W) && t->Neighbor(E) );
	++cnt; // N
	if( t->n() == NW ){
	  if( tcp == t->Neighbor(W)->CP() ) ++cnt; // NW
	  if( tcp == t->Neighbor(E)->CP() ) ++cnt; // NE
	}
      }
      if( t->Neighbor(W) && !t->Neighbor(W)->CompleteSetOfChildren() &&
	  t->nw() && t->sw() && ( t->nw()->Leaf() || t->sw()->Leaf() ) && 
	  tcp >= t->Neighbor(W)->CP() ){
	assert( t->Neighbor(N) && t->Neighbor(S) );
	++cnt; // W
	if( t->n() == NW ){
	  if( tcp == t->Neighbor(N)->CP() ) ++cnt; // NW
	  if( tcp == t->Neighbor(S)->CP() ) ++cnt; // SW
	}
      }
      if( t->Neighbor(S) && !t->Neighbor(S)->CompleteSetOfChildren() &&
	  t->sw() && t->se() && ( t->sw()->Leaf() || t->se()->Leaf() ) && 
	  tcp >= t->Neighbor(S)->CP() ){
	assert( t->Neighbor(W) && t->Neighbor(E) );
	++cnt; // S
	if( t->n() == SW ){
	  if( tcp == t->Neighbor(W)->CP() ) ++cnt; // SW
	  if( tcp == t->Neighbor(E)->CP() ) ++cnt; // SE
	}
      }
      if( t->Neighbor(E) && !t->Neighbor(E)->CompleteSetOfChildren() &&
	  t->ne() && t->se() && ( t->ne()->Leaf() || t->se()->Leaf() ) && 
	  tcp >= t->Neighbor(E)->CP() ){
	assert( t->Neighbor(N) && t->Neighbor(S) );
	++cnt; // E
	if( t->n() == SE ){
	  if( tcp == t->Neighbor(N)->CP() ) ++cnt; // NE
	  if( tcp == t->Neighbor(S)->CP() ) ++cnt; // SE
	}
      }
    }
  } 
  unsigned long cnt; 
};

// write in the "smf" format
class WriteVertex : public unary_function<CVec*, void>{
public:
  WriteVertex( ostream* os, std::map<const CVec*, unsigned int>* vmap, unsigned int seq ) : out( os ), map( vmap ), i( seq ) {}
  virtual void operator()( CVec *v ) {
    *out << *v << endl; assert( map->find(v) == map->end() );
    (*map)[v] = i++;
  }
  ostream* out;
  std::map<const CVec*, unsigned int>* map;
  unsigned int i;
};

// write in the "iv" format
class WriteVertexIV : public WriteVertex{
public:
  WriteVertexIV( ostream* os, std::map<const CVec*, unsigned int>* vmap, unsigned int seq, const char * space ) : WriteVertex( os, vmap, seq ), indent( space ) {}
  virtual void operator()( CVec *v ) {
    *out << indent << *v << "," << endl; assert( map->find(v) == map->end() );
    (*map)[v] = i++;
  }
  const char * indent;
};

// write in the "smf" format
class WriteQuads : public unary_function<QTree*, void>{
public:
  WriteQuads( ostream* os, std::map<const CVec*, unsigned int>* vmap ) 
    : out(os), map( vmap ) {}
  virtual void QT ( const QTree *nw, const QTree *sw, const QTree *se, const QTree *ne ){
    assert( map->find( &nw->loc() ) != map->end());
    assert( map->find( &sw->loc() ) != map->end());
    assert( map->find( &se->loc() ) != map->end());
    assert( map->find( &ne->loc() ) != map->end());
    *out << (*map)[&nw->loc()] << " "
	 << (*map)[&sw->loc()] << " "
	 << (*map)[&se->loc()] << " "
	 << (*map)[&ne->loc()] << " " << END_FACE_INDEX << endl;
  }
  void operator() ( const QTree *t ) {
    if( t->CompleteSetOfChildren() ){
      if( !t->nw()->CompleteSetOfChildren() ||
	  !t->sw()->CompleteSetOfChildren() ||
	  !t->se()->CompleteSetOfChildren() ||
	  !t->ne()->CompleteSetOfChildren() ){
	QT( t->nw(), t->sw(), t->se(), t->ne() ); // the center guy
	const Cp *tcp = t->CP();
	const Cp *tncp = t->Neighbor(N)->CP();
	const Cp *twcp = t->Neighbor(W)->CP();
	const Cp *tscp = t->Neighbor(S)->CP();
	const Cp *tecp = t->Neighbor(E)->CP();
	switch( t->n() ){
	case NW:
	  if( tcp >= tecp || !t->Neighbor(E)->CompleteSetOfChildren() ){ // E
	    const QTree * se = t->se()->Neighbor(E); assert( se->Depth() > t->Depth() );
	    const QTree * ne = t->ne()->Neighbor(E); assert( ne->Depth() > t->Depth() );
	  QT( t->ne(), t->se(), se, ne );
	  }
	  if( tcp >= tscp || !t->Neighbor(S)->CompleteSetOfChildren() ){ // S
	    const QTree * sw = t->sw()->Neighbor(S); assert( sw->Depth() > t->Depth() );
	    const QTree * se = t->se()->Neighbor(S); assert( se->Depth() > t->Depth() );
	    QT( t->sw(), sw, se, t->se() );
	  }
	  if( tcp >= tncp || !t->Neighbor(N)->CompleteSetOfChildren() ){ // N
	    const QTree * nw = t->nw()->Neighbor(N); assert( nw->Depth() > t->Depth() );
	    const QTree * ne = t->ne()->Neighbor(N); assert( ne->Depth() > t->Depth() );
	    QT( nw, t->nw(), t->ne(), ne );
	    if( !t->Root() ){ // NE
	      const QTree * nw = t->ne()->Neighbor(N); assert( nw->Depth() > t->Depth() );
	      const QTree * se = t->ne()->Neighbor(E); assert( se->Depth() > t->Depth() );
	      const QTree * ne = se->Neighbor(N); assert( ne->Depth() > t->Depth() );
	      QT( nw, t->ne(), se, ne );
	    }
	  }
	  if( tcp >= twcp || !t->Neighbor(W)->CompleteSetOfChildren() ){ // W
	    const QTree * nw = t->nw()->Neighbor(W); assert( nw->Depth() > t->Depth() );
	    const QTree * sw = t->sw()->Neighbor(W); assert( sw->Depth() > t->Depth() );
	    QT( nw, sw, t->sw(), t->nw() );
	    if( !t->Root() ){ // SW
	      const QTree * nw = t->sw()->Neighbor(W); assert( nw->Depth() > t->Depth() );
	      const QTree * se = t->sw()->Neighbor(S); assert( se->Depth() > t->Depth() );
	      const QTree * sw = se->Neighbor(W); assert( sw->Depth() > t->Depth() );
	      QT( nw, sw, se, t->sw() );
	    }
	  }
	  if( ( tcp == tncp && ( tcp >= twcp || !t->Neighbor(W)->CompleteSetOfChildren()))
	      || ( tcp == twcp && ( tcp >= tncp || !t->Neighbor(N)->CompleteSetOfChildren()))){ 
	    if( !t->Root() ){ // NW
	      const QTree * ne = t->nw()->Neighbor(N); assert( ne->Depth() > t->Depth() );
	      const QTree * sw = t->nw()->Neighbor(W); assert( sw->Depth() > t->Depth() );
	      const QTree * nw = ( tcp == tncp ) ? ne->Neighbor(W) : sw->Neighbor(N); assert( nw->Depth() > t->Depth() ); 
	      QT( nw, sw, t->nw(), ne );
	    }
	  }
	  if( !t->Root() ){// SE
	    const QTree * sw = t->se()->Neighbor(S); assert( sw->Depth() > t->Depth() );
	    const QTree * se = sw->Neighbor(E); assert( sw->Depth() > t->Depth() );
	    const QTree * ne = t->se()->Neighbor(E); assert( ne->Depth() > t->Depth() );
	    QT( t->se(), sw, se, ne );
	  }
	  break;
	case SW:
	  { // E
	    const QTree * se = t->se()->Neighbor(E); assert( se->Depth() > t->Depth() );
	    const QTree * ne = t->ne()->Neighbor(E); assert( ne->Depth() > t->Depth() );
	    QT( t->ne(), t->se(), se, ne );
	  }
	  if( tcp >= twcp || !t->Neighbor(W)->CompleteSetOfChildren() ){ // W
	    const QTree *nw = t->nw()->Neighbor(W); assert( nw->Depth() > t->Depth() );
	    const QTree *sw = t->sw()->Neighbor(W); assert( sw->Depth() > t->Depth() );
	    QT( nw, sw, t->sw(), t->nw() );
	  }
	  if( tcp > tscp || !t->Neighbor(S)->CompleteSetOfChildren() ){ // S, SE
	    // S
	    const QTree *sw = t->sw()->Neighbor(S); assert( sw->Depth() > t->Depth() );
	    const QTree *se = t->se()->Neighbor(S); assert( se->Depth() > t->Depth() );
	    QT( t->sw(), sw, se, t->se() );
	    // SE
	    const QTree *sw2 = t->se()->Neighbor(S); assert( sw2->Depth() > t->Depth() );
	    const QTree *ne2 = t->se()->Neighbor(E); assert( ne2->Depth() > t->Depth() ); 
	    const QTree *se2 = ne2->Neighbor(S); assert( se2->Depth() > t->Depth() );
	    QT( t->se(), sw2, se2, ne2 );
	    if( tcp == twcp ){ // SW
	      const QTree * nw = t->sw()->Neighbor(W); assert( nw->Depth() > t->Depth() );
	      const QTree * se = t->sw()->Neighbor(S); assert( se->Depth() > t->Depth() );
	      const QTree * sw = nw->Neighbor(S); assert( sw->Depth() > t->Depth() );
	      QT( nw, sw, se, t->sw() );
	    }
	  }
	  break;
	case SE:
	  { // N
	    const QTree * nw = t->nw()->Neighbor(N); assert( nw->Depth() > t->Depth() );
	    const QTree * ne = t->ne()->Neighbor(N); assert( ne->Depth() > t->Depth() );
	    QT( nw, t->nw(), t->ne(), ne );
	  }
	  if( tcp >= tscp || !t->Neighbor(S)->CompleteSetOfChildren() ){ // S
	    const QTree *sw = t->sw()->Neighbor(S); assert( sw->Depth() > t->Depth() );
	    const QTree *se = t->se()->Neighbor(S); assert( se->Depth() > t->Depth() );
	    QT( t->sw(), sw, se, t->se() );
	  }
	  if( tcp > tecp || !t->Neighbor(E)->CompleteSetOfChildren() ){ // E, NE
	    // E
	    const QTree *se = t->se()->Neighbor(E); assert( se->Depth() > t->Depth() );
	    const QTree *ne = t->ne()->Neighbor(E); assert( ne->Depth() > t->Depth() );
	    QT( t->ne(), t->se(), se, ne );
	    // NE
	    const QTree * nw2 = t->ne()->Neighbor(N); assert( nw2->Depth() > t->Depth() );
	    const QTree * se2 = t->ne()->Neighbor(E); assert( se2->Depth() > t->Depth() );
	    const QTree * ne2 = nw2->Neighbor(E); assert( ne2->Depth() > t->Depth() );
	    QT( nw2, t->ne(), se2, ne2 );
	    if( tcp == tscp ){ // SE
	      const QTree *sw = t->se()->Neighbor(S); assert( sw->Depth() > t->Depth() );
	      const QTree *ne = t->se()->Neighbor(E); assert( ne->Depth() > t->Depth() );
	      const QTree *se = sw->Neighbor(E); assert( se->Depth() > t->Depth() );
	      QT( t->se(), sw, se, ne );
	    }
	  }
	  break;
	case NE:
	  if( tcp > tncp || !t->Neighbor(N)->CompleteSetOfChildren() ){ // N
	    const QTree * nw = t->nw()->Neighbor(N); assert( nw->Depth() > t->Depth() );
	    const QTree * ne = t->ne()->Neighbor(N); assert( ne->Depth() > t->Depth() );
	    QT( nw, t->nw(), t->ne(), ne );
	  }
	  if( tcp > tecp || !t->Neighbor(E)->CompleteSetOfChildren() ){ // E
	    const QTree *se = t->se()->Neighbor(E); assert( se->Depth() > t->Depth() );
	    const QTree *ne = t->ne()->Neighbor(E); assert( ne->Depth() > t->Depth() );
	    QT( t->ne(), t->se(), se, ne );
	  }
	  break;
	default: die();
	}
      }
    }
    else{
      // deal with the special case on a boundary non-extreme vertex
      // if some of its neighbors are boundary non-extreme vertices

      const Cp * tcp = t->CP();

      if( t->Neighbor(N) && !t->Neighbor(N)->CompleteSetOfChildren() &&
	  t->nw() && t->ne() && ( t->nw()->Leaf() || t->ne()->Leaf() ) && 
	  tcp >= t->Neighbor(N)->CP() ){
	assert( t->Neighbor(W) && t->Neighbor(E) );
	// N
	const QTree * nw = t->nw()->Neighbor(N); assert( nw->Depth() > t->Depth() );
	const QTree * ne = t->ne()->Neighbor(N); assert( ne->Depth() > t->Depth() );
	QT( nw, t->nw(), t->ne(), ne );

	if( t->n() == NW ){
	  if( tcp == t->Neighbor(W)->CP() ){ // NW
	    const QTree * ne = t->nw()->Neighbor(N); assert( ne->Depth() > t->Depth() );
	    const QTree * sw = t->nw()->Neighbor(W); assert( sw->Depth() > t->Depth() );
	    const QTree * nw = sw->Neighbor(N); assert( nw->Depth() > t->Depth() ); 
	    QT( nw, sw, t->nw(), ne );
	  }
	  if( tcp == t->Neighbor(E)->CP() ){ // NE
	    const QTree * nw = t->ne()->Neighbor(N); assert( nw->Depth() > t->Depth() );
	    const QTree * se = t->ne()->Neighbor(E); assert( se->Depth() > t->Depth() );
	    const QTree * ne = se->Neighbor(N); assert( ne->Depth() > t->Depth() );
	    QT( nw, t->ne(), se, ne );
	  }
	}
      }

      if( t->Neighbor(W) && !t->Neighbor(W)->CompleteSetOfChildren() &&
	  t->nw() && t->sw() && ( t->nw()->Leaf() || t->sw()->Leaf() ) && 
	  tcp >= t->Neighbor(W)->CP() ){
	assert( t->Neighbor(N) && t->Neighbor(S) );
	// W
	const QTree * nw = t->nw()->Neighbor(W); assert( nw->Depth() > t->Depth() );
	const QTree * sw = t->sw()->Neighbor(W); assert( sw->Depth() > t->Depth() );
	QT( nw, sw, t->sw(), t->nw() );

	if( t->n() == NW ){
	  if( tcp == t->Neighbor(N)->CP() ){ // NW
	    const QTree * ne = t->nw()->Neighbor(N); assert( ne->Depth() > t->Depth() );
	    const QTree * sw = t->nw()->Neighbor(W); assert( sw->Depth() > t->Depth() );
	    const QTree * nw = ne->Neighbor(W); assert( nw->Depth() > t->Depth() ); 
	    QT( nw, sw, t->nw(), ne );
	  }
	  if( tcp == t->Neighbor(S)->CP() ){ // SW
	    const QTree * nw = t->sw()->Neighbor(W); assert( nw->Depth() > t->Depth() );
	    const QTree * se = t->sw()->Neighbor(S); assert( se->Depth() > t->Depth() );
	    const QTree * sw = se->Neighbor(W); assert( sw->Depth() > t->Depth() );
	    QT( nw, sw, se, t->sw() );
	  }
	}
      }

      if( t->Neighbor(S) && !t->Neighbor(S)->CompleteSetOfChildren() &&
	  t->sw() && t->se() && ( t->sw()->Leaf() || t->se()->Leaf() ) && 
	  tcp >= t->Neighbor(S)->CP() ){
	assert( t->Neighbor(W) && t->Neighbor(E) );
	// S
	const QTree * sw = t->sw()->Neighbor(S); assert( sw->Depth() > t->Depth() );
	const QTree * se = t->se()->Neighbor(S); assert( se->Depth() > t->Depth() );
	QT( t->sw(), sw, se, t->se() );

	if( t->n() == SW ){
	  if( tcp == t->Neighbor(W)->CP() ){ // SW
	    const QTree * nw = t->sw()->Neighbor(W); assert( nw->Depth() > t->Depth() );
	    const QTree * se = t->sw()->Neighbor(S); assert( se->Depth() > t->Depth() );
	    const QTree * sw = nw->Neighbor(S); assert( sw->Depth() > t->Depth() );
	    QT( nw, sw, se, t->sw() );
	  }
	  if( tcp == t->Neighbor(E)->CP() ){ // SE
	    const QTree *sw = t->se()->Neighbor(S); assert( sw->Depth() > t->Depth() );
	    const QTree *ne = t->se()->Neighbor(E); assert( ne->Depth() > t->Depth() ); 
	    const QTree *se = ne->Neighbor(S); assert( se->Depth() > t->Depth() );
	    QT( t->se(), sw, se, ne );
	  }
	}
      }

      if( t->Neighbor(E) && !t->Neighbor(E)->CompleteSetOfChildren() &&
	  t->ne() && t->se() && ( t->ne()->Leaf() || t->se()->Leaf() ) && 
	  tcp >= t->Neighbor(E)->CP() ){
	assert( t->Neighbor(N) && t->Neighbor(S) );
	// E
	const QTree *se = t->se()->Neighbor(E); assert( se->Depth() > t->Depth() );
	const QTree *ne = t->ne()->Neighbor(E); assert( ne->Depth() > t->Depth() );
	QT( t->ne(), t->se(), se, ne );

	if( t->n() == SE ){
	  if( tcp == t->Neighbor(N)->CP() ){ // NE
	    const QTree * nw = t->ne()->Neighbor(N); assert( nw->Depth() > t->Depth() );
	    const QTree * se = t->ne()->Neighbor(E); assert( se->Depth() > t->Depth() );
	    const QTree * ne = nw->Neighbor(E); assert( ne->Depth() > t->Depth() );
	    QT( nw, t->ne(), se, ne );
	  }
	  if( tcp == t->Neighbor(S)->CP() ){ // SE
	    const QTree *sw = t->se()->Neighbor(S); assert( sw->Depth() > t->Depth() );
	    const QTree *ne = t->se()->Neighbor(E); assert( ne->Depth() > t->Depth() );
	    const QTree *se = sw->Neighbor(E); assert( se->Depth() > t->Depth() );
	    QT( t->se(), sw, se, ne );
	  }
	}
      }
    }
  }
  ostream* out;
  std::map<const CVec*, unsigned int>* map;
};

// write in the "iv" format
class WriteQuadsIV : public WriteQuads{
public:
  WriteQuadsIV( ostream* os, std::map<const CVec*, unsigned int>* vmap, const char * space ) : WriteQuads( os, vmap ), indent( space ) {}
  virtual void QT( const QTree *nw, const QTree *sw, const QTree *se, const QTree *ne ) {
    assert( map->find( &nw->loc() ) != map->end());
    assert( map->find( &sw->loc() ) != map->end());
    assert( map->find( &se->loc() ) != map->end());
    assert( map->find( &ne->loc() ) != map->end());
    *out << indent;
    *out << (*map)[&nw->loc()] << ", "
	 << (*map)[&sw->loc()] << ", "
	 << (*map)[&se->loc()] << ", "
	 << (*map)[&ne->loc()] << ", " << END_FACE_INDEX << "," << endl;
  }
  const char * indent;
};

// write in the "smf" format
class WriteTriangle : public unary_function<QTree*, void>{
public:
  WriteTriangle( ostream* os, std::map<const CVec*, unsigned int>* vmap )
    : out( os ), map( vmap ) {}
  virtual void DT ( const unsigned int a, const unsigned int b, const unsigned int c ){
    assert( a != 0xffffffff && b != 0xffffffff && c != 0xffffffff );
    *out << a << " " << b << " " << c << " " << END_FACE_INDEX << endl;
  }
  void Rot( const int amount,
	    unsigned int &a,
	    unsigned int &b,
	    unsigned int &c,
	    unsigned int &d ) const{
    unsigned int tmp;
    switch( amount ){
    case 1: case -3: tmp = a; a = b; b = c; c = d; d = tmp; break;
    case -2: case 2: tmp = a; a = b; b = c; c = d; d = tmp; tmp = a; a = b; b = c; c = d; d = tmp; break;
    case 3: case -1: tmp = d; d = c; c = b; b = a; a = tmp; break;
    case 0: break;
    default: die();
    }
  }
  void operator() ( QTree *t ) {
    if( t->Leaf() && t->Corner(NW) && t->Corner(SW) && t->Corner(SE) && t->Corner(NE) ){
      Perms np;
      int config = 0;
      const QTree *n;
      assert( map->find(t->Corner(NW)) != map->end() );
      unsigned int nwi = (*map)[t->Corner(NW)];
      assert( map->find(t->Corner(SW)) != map->end() );
      unsigned int swi = (*map)[t->Corner(SW)];
      assert( map->find(t->Corner(SE)) != map->end() );
      unsigned int sei = (*map)[t->Corner(SE)];
      assert( map->find(t->Corner(NE)) != map->end() );
      unsigned int nei = (*map)[t->Corner(NE)];

      // west, south, east, north - edge index
      unsigned int wi = 0xffffffff, si = 0xffffffff, ei = 0xffffffff, ni = 0xffffffff;
      if( ( n = t->Neighbor( W, &np ) ) &&
	  ( n->c(np[NE]) && n->c(np[SE]) ) ){
	assert( map->find(n->c(np[NE])->Corner(np[SE])) != map->end() );
	wi = (*map)[n->c(np[NE])->Corner(np[SE])]; config |= 1;
      }
      if( ( n = t->Neighbor( S, &np ) ) &&
	  ( n->c(np[NW]) && n->c(np[NE]) ) ){
	assert( map->find(n->c(np[NW])->Corner(np[NE])) != map->end() );
	si = (*map)[n->c(np[NW])->Corner(np[NE])]; config |= 2;
      }
      if( ( n = t->Neighbor( E, &np ) ) &&
	  ( n->c(np[NW]) && n->c(np[SW]) ) ){
	assert( map->find(n->c(np[NW])->Corner(np[SW])) != map->end() );
	ei = (*map)[n->c(np[NW])->Corner(np[SW])]; config |= 4;
      }
      if( ( n = t->Neighbor( N, &np ) ) &&
	  ( n->c(np[SW]) && n->c(np[SE]) ) ){
	assert( map->find(n->c(np[SW])->Corner(np[SE])) != map->end() );
	ni = (*map)[n->c(np[SW])->Corner(np[SE])]; config |= 8;
      }

      // rotate names
      switch( t->n() ){
      case NW: break;
      case SW:
	Rot( 1, nwi, swi, sei, nei ); Rot( 1, wi, si, ei, ni );
	config = (config>>1) | ((config&0x1)<<3); break;
      case SE:
	Rot( 2, nwi, swi, sei, nei ); Rot( 2, wi, si, ei, ni );
	config = (config>>2) | ((config&0x3)<<2); break;
      case NE:
	Rot( 3, nwi, swi, sei, nei ); Rot( 3, wi, si, ei, ni );
        config = (config>>3) | ((config&0x7)<<1); break;
      default: die();
      }

      // now everything looks like it is NW
      switch( config ){
      case 0: DT(nwi,swi,sei); DT(sei,nei,nwi); break;
      case 1: DT(nwi,wi,nei); DT(wi,sei,nei); DT(wi,swi,sei); break;
      case 2: DT(nwi,swi,si); DT(si,nei,nwi); DT(si,sei,nei); break;
      case 4: DT(swi,sei,ei); DT(ei,nwi,swi); DT(ei,nei,nwi); break;
      case 8: DT(nwi,swi,ni); DT(swi,sei,ni); DT(sei,nei,ni); break;
      case 3: DT(wi,swi,si); DT(si,nei,wi); DT(si,sei,nei); DT(nei,nwi,wi); break;
      case 6: DT(nwi,swi,si); DT(si,ei,nwi); DT(si,sei,ei); DT(ei,nei,nwi); break;
      case 12:DT(ni,nwi,swi); DT(swi,ei,ni); DT(swi,sei,ei); DT(ei,nei,ni); break;
      case 9: DT(nwi,wi,ni); DT(wi,swi,sei); DT(sei,ni,wi); DT(sei,nei,ni); break;
      case 7: DT(nwi,wi,si); DT(wi,swi,si); DT(si,sei,ei); DT(ei,nei,si); DT(nei,nwi,si); break;
      case 14:DT(swi,si,ei); DT(si,sei,ei); DT(ei,nei,ni); DT(ei,ni,nwi); DT(ei,nwi,swi); break;
      case 13:DT(ni,nwi,wi); DT(ni,wi,swi); DT(ni,swi,sei); DT(ni,sei,ei); DT(ni,ei,nei); break;
      case 11:DT(wi,swi,si); DT(wi,si,sei); DT(wi,sei,nei); DT(wi,nei,ni); DT(wi,ni,nwi); break;
      case 5: DT(wi,swi,sei); DT(wi,sei,ei); DT(wi,ei,nei); DT(wi,nei,nwi); break;
      case 10:DT(ni,nwi,swi); DT(ni,swi,si); DT(ni,si,sei); DT(ni,sei,nei); break;
      case 15:DT(ni,nwi,wi); DT(wi,swi,si); DT(si,sei,ei); DT(ei,nei,ni); DT(wi,ei,ni); DT(wi,si,ei); break;
      default: die();
      }
    }
  }
  ostream* out;
  std::map<const CVec*, unsigned int>* map;
};

// write in the "iv" format
class WriteTriangleIV : public WriteTriangle{
public:
  WriteTriangleIV( ostream* os, std::map<const CVec*, unsigned int>* vmap, const char * space ) : WriteTriangle( os, vmap ), indent( space ) {}
  virtual void DT ( const unsigned int a, const unsigned int b, const unsigned int c ){
    assert( a != 0xffffffff && b != 0xffffffff && c != 0xffffffff );
    *out << indent << a << ", " << b << ", " << c << ", " << END_FACE_INDEX << "," << endl;
  }
  const char * indent;
};

// output the triangle mesh in the "smf" format
// vertices are corners of QTrees
ostream&
CMesh::WriteDualTriangle( ostream &os )
{
  // this puts out the dual graph as triangles
  // first count all triangles and all corner vertices and initialize them
  // to zero
  const unsigned int fno = m_fc.size();
  unsigned int vno = fno;
  unsigned int i;
  const CVec Zero(0,0,0);
  for( i = 0; i < vno; ++i ) *m_fc[i]->c() = Zero;
  std::vector<Cp*>::const_iterator cpb = m_cpc.begin(), cpi = cpb, cpe = m_cpc.end();
  for( ; cpi != cpe; ++cpi ) vno += (*cpi)->Root()->ForEachOddVertex( CountVertices() ).cnt;

  // count triangles and contributions to corner vertices
  unsigned int tno = 0;
  for( cpi = cpb; cpi != cpe; ++cpi ) tno += (*cpi)->Root()->ForEachQTree( CountTriangles() ).cnt;

  // the centroids of global faces may need renormalization now because the
  // number of incident dual faces may not be 4
  for( i = 0; i < fno; ++i ) *m_fc[i]->c() *= real(4)/m_fc[i]->CPs().size();
  

  if( !( os << vno << " " << tno << endl ) ) return os;

  std::map<const CVec*, unsigned int> vmap;
  std::vector<CFace*>::const_iterator vi = m_fc.begin(), ve = m_fc.end();
  os.precision(16);
  for( i = 0; vi != ve; ++vi, ++i ){
    if( !( os << *(*vi)->c() << endl ) ) return os;
    vmap[(*vi)->c()] = i;
  }
  // i gets carried along as the increaing sequence number for the hash map
  for( cpi = cpb; cpi != cpe; ++cpi ) i = (*cpi)->Root()->ForEachOddVertex( WriteVertex( &os, &vmap, i ) ).i;

  assert( i == vno );
  // all vertices have been written and have been entered into the hash
  // map
  // now write the triangles
  for( cpi = cpb; cpi != cpe; ++cpi ) (*cpi)->Root()->ForEachQTree( WriteTriangle( &os, &vmap ) );

  return os;
}

// output triangle mesh in the "iv" format
// vertices are the corners of QTrees
ostream&
CMesh::WriteDualTriangleIV( ostream& os ){
  // this puts out the dual graph as triangles
  // first count all triangles and all corner vertices and initialize them
  // to zero
  const unsigned int fno = m_fc.size();
  unsigned int vno = fno;
  unsigned int i;
  const CVec Zero(0,0,0);
  for( i = 0; i < vno; ++i ) *m_fc[i]->c() = Zero;
  std::vector<Cp*>::const_iterator cpb = m_cpc.begin(), cpi = cpb, cpe = m_cpc.end();
  for( ; cpi != cpe; ++cpi ) vno += (*cpi)->Root()->ForEachOddVertex( CountVertices() ).cnt;

  // count triangles and contributions to corner vertices
  unsigned int tno = 0;
  for( cpi = cpb; cpi != cpe; ++cpi ) tno += (*cpi)->Root()->ForEachQTree( CountTriangles() ).cnt;

  // the centroids of global faces may need renormalization now because the
  // number of incident dual faces may not be 4
  for( i = 0; i < fno; ++i ) *m_fc[i]->c() *= real(4)/m_fc[i]->CPs().size();
  
  if( !( os << IV_HEADER << endl << endl ) ) return os;

  if( !( os << "Separator {" << endl ) ) return os;
  if( !( os << INDENT1 << "Coordinate3 {" << endl ) ) return os;
  if( !( os << INDENT2 << "point [" << endl ) ) return os;

  std::map<const CVec*, unsigned int> vmap;
  std::vector<CFace*>::const_iterator vi = m_fc.begin(), ve = m_fc.end();
  os.precision(16);
  for( i = 0; vi != ve; ++vi, ++i ){
    if( !( os << INDENT3 << *(*vi)->c() << "," << endl ) ) return os;
    vmap[(*vi)->c()] = i;
  }
  // i gets carried along as the increaing sequence number for the hash map
  for( cpi = cpb; cpi != cpe; ++cpi ) i = (*cpi)->Root()->ForEachOddVertex( WriteVertexIV( &os, &vmap, i, INDENT3 ) ).i;

  assert( i == vno );

  if( !( os << INDENT2 << "]" << endl ) ) return os; // point [
  if( !( os << INDENT1 << "}" << endl ) ) return os; // Coordinate3 {
  if( !( os << INDENT1 << "IndexedFaceSet {" << endl ) ) return os;
  if( !( os << INDENT2 << "coordIndex [" << endl ) ) return os;

  // all vertices have been written and have been entered into the hash map
  // now write the triangles
  for( cpi = cpb; cpi != cpe; ++cpi ) (*cpi)->Root()->ForEachQTree( WriteTriangleIV( &os, &vmap, INDENT3 ) );

  if( !( os << INDENT2 << "]" << endl ) ) return os; // coordIndex [
  if( !( os << INDENT1 << "}" << endl ) ) return os; // IndexedFaceSet {

  os << "}" << endl; // Separator {
  return os;
}

// write in the "smf" format
// output the PrimalPolygons shared by more than 3 Cps
static void
WriteFace( ostream& os, std::map<const CVec*, unsigned int>* vmap, CFace *f )
{
  // for all incident cp's burrow to the bottom
  const unsigned int size = f->CPs().size();
  unsigned int i;
  for( i = 0; i < size; ++i ){
    const Cp *cp = f->CPs()[i];
    const QTree *t = cp->Root();
    const QTreeName qtn = t->Corner( f->c() );
    const CVec *v = &t->loc();
    // descend
    while( t->c( qtn ) ){ t = t->c( qtn ); v = &t->loc(); }
    assert( vmap->find(v) != vmap->end() );
    os << (*vmap)[v] << " ";
  }
  os << END_FACE_INDEX << endl;
}

// write in the "iv" format
// output the PrimalPolygons shared by more than 3 Cps
static void
WriteFaceIV( ostream& os, std::map<const CVec*, unsigned int>* vmap, CFace *f )
{
  // for all incident cp's burrow to the bottom
  const unsigned int size = f->CPs().size();
  unsigned int i;
  if( size > 0 ) os << INDENT3;
  for( i = 0; i < size; ++i ){
    const Cp *cp = f->CPs()[i];
    const QTree *t = cp->Root();
    const QTreeName qtn = t->Corner( f->c() );
    const CVec *v = &t->loc();
    // descend
    while( t->c( qtn ) ){ t = t->c( qtn ); v = &t->loc(); }
    assert( vmap->find(v) != vmap->end() );
    os << (*vmap)[v] << ", ";
  }
  os << END_FACE_INDEX << "," << endl;
}

// write the polygonal mesh in the "smf" format
// vertices are QTrees
ostream&
CMesh::WritePrimalPolygon( ostream &os )
{
  // each qtree leaf has a vertex
  unsigned int vno = 0;
  std::vector<Cp*>::const_iterator cpb = m_cpc.begin(), cpi = cpb, cpe = m_cpc.end();
  // count vertices
  for( cpi = cpb; cpi != cpe; ++cpi ) vno += (*cpi)->Root()->ForEachQTree( CountLeafCPs() ).cnt;

  // now count polygons; there is one for each CFace
  unsigned int pno = m_fc.size();
  for( cpi = cpb; cpi != cpe; ++cpi ) pno += (*cpi)->Root()->ForEachQTree( CountQuads() ).cnt;

  if( !( os << vno << " " << pno << endl ) ) return os;

  std::map<const CVec*, unsigned int> vmap;
  unsigned int i;
  os.precision(16);
  for( i = 0, cpi = cpb; cpi != cpe; ++cpi ) i = (*cpi)->Root()->ForEachQTree( WriteLeafCPs( &os, &vmap, i) ).i;

  assert( i == vno );
  
  // write all extraordinary faces first
  std::vector<CFace*>::const_iterator vi = m_fc.begin(), ve = m_fc.end();
  for( ; vi != ve; ++vi ) WriteFace( os, &vmap, *vi );

  // i gets carried along as the increaing sequence number for the hash map
  for( cpi = cpb; cpi != cpe; ++cpi ) (*cpi)->Root()->ForEachQTree( WriteQuads( &os, &vmap ) );

  return os;
}

// write the polygonal mesh in the "iv" format
// vertices are QTrees
ostream&
CMesh::WritePrimalPolygonIV( ostream& os )
{
  // each qtree leaf has a vertex
  unsigned int vno = 0;
  std::vector<Cp*>::const_iterator cpb = m_cpc.begin(), cpi = cpb, cpe = m_cpc.end();
  // count vertices
  for( cpi = cpb; cpi != cpe; ++cpi ) vno += (*cpi)->Root()->ForEachQTree( CountLeafCPs() ).cnt;

  // now count polygons; there is one for each CFace
  unsigned int pno = m_fc.size();
  for( cpi = cpb; cpi != cpe; ++cpi ) pno += (*cpi)->Root()->ForEachQTree( CountQuads() ).cnt;

  if( !( os << IV_HEADER << endl << endl ) ) return os;
  
  if( !( os << "Separator {" << endl ) ) return os;
  if( !( os << INDENT1 << "Coordinate3 {" << endl ) ) return os;
  if( !( os << INDENT2 << "point [" << endl ) ) return os;

  std::map<const CVec*, unsigned int> vmap;
  unsigned int i;
  os.precision(16);
  for( i = 0, cpi = cpb; cpi != cpe; ++cpi ) i = (*cpi)->Root()->ForEachQTree( WriteLeafCPsIV( &os, &vmap, i, INDENT3 ) ).i;

  assert( i == vno );
  
  if( !( os << INDENT2 << "]" << endl ) ) return os; // point [
  if( !( os << INDENT1 << "}" << endl ) ) return os; // Coordinate3 {
  if( !( os << INDENT1 << "IndexedFaceSet {" << endl ) ) return os;
  if( !( os << INDENT2 << "coordIndex [" << endl ) ) return os;

  // write all extraordinary faces first
  std::vector<CFace*>::const_iterator vi = m_fc.begin(), ve = m_fc.end();
  for( ; vi != ve; ++vi ) WriteFaceIV( os, &vmap, *vi );

  // i gets carried along as the increaing sequence number for the hash map
  for( cpi = cpb; cpi != cpe; ++cpi ) (*cpi)->Root()->ForEachQTree( WriteQuadsIV( &os, &vmap, INDENT3 ) );

  if( !( os << INDENT2 << "]" << endl ) ) return os; // coordIndex [
  if( !( os << INDENT1 << "}" << endl ) ) return os; // IndexedFaceSet {

  os << "}" << endl; // Separator {

  return os;
}
