// -*- Mode: c++ -*-
// $Id: OneRing.cpp,v 1.7 2000/04/28 22:14:46 ps Exp $
// $Source: /ufs/multires/projects//src/CourseQuad/OneRing.cpp,v $

/**
 * Copyright 1995-2000, Peter Schr\"oder, California Institute of
 * Technology (Caltech) 
 *
 * April 26, 2000
 * All Rights Reserved
 * 
 * The use of this software is permitted for non-commercial, educational,
 * and research use only. The software and/or related materials are
 * provided "as-is" without warranty of any kind including any warranties
 * of performance or merchantability or fitness for a particular use or
 * purpose (as set forth in UCC 23212-2313) or for any purpose
 * whatsoever, for the licensed product, however used. In no event shall
 * Caltech be liable for any damages and/or costs, including but not
 * limited to incidental or consequential damages of any kind, including
 * economic damage or injury to property and lost profits, regardless of
 * whether Caltech shall be advised, have reason to know, or in fact
 * shall know of the possibility. User bears all risk relating to quality
 * and performance of the software and/or related materials.
 * 
 * Any use other than non-commercial, educational, or research, or any
 * redistribution in original or modified form requires prior written
 * authorization from the copyright holder.
 */

#include <math.h>
#include <algorithm>
#include "real.h"
#include "util.h"
#include "cvec.h"
#include "Cp.h"
#include "OneRing.h"
#include "QTree.h"
#include "Perms.h"

bool OneRing::m_init = false;
real OneRing::m_ds[__MAXFACESIZE][__MAXFACESIZE];
real OneRing::m_me[__MAXFACESIZE][__MAXFACESIZE];
real OneRing::m_bq[4][__MAXFACESIZE][__MAXFACESIZE];

// fixed coefficients in BQ scheme
real OneRing::m_fbq[4][8] =
// center, N, W, NE, SW, E, S, SE
{{real(51)/256,real(15)/256,real(15)/256,real(5)/256,real(5)/256,real(10)/256,real(10)/256,real(1)/256},
 {real(79)/256,real(3)/256,real(35)/256,real(1)/256,real(25)/256,real(10)/256,real(50)/256,real(5)/256},
 {real(91)/256,real(7)/256,real(7)/256,real(5)/256,real(5)/256,real(50)/256,real(50)/256,real(25)/256},
 {real(79)/256,real(35)/256,real(3)/256,real(25)/256,real(1)/256,real(50)/256,real(10)/256,real(5)/256}};

void
OneRing::BuildTables( void ) const
{
  for( int i = 3; i < __MAXFACESIZE; ++i ){
    const real tk = 2 * real(M_PI) / i;
    int j;
    for( j = 0; j < i; ++j ){
      // original DS rule
      // m_ds[i][j] = ( 3 + 2 * real( cos( tk * j ) ) ) / ( 4 * i );
      // DS rule of Catmull-Clark (pure repeated averaging)
      m_ds[i][j] = real(1) / ( 4 * i );
      m_me[i][j] = ( 1 + real( cos( tk * j ) ) ) / i;
    }
    // original DS rule
    // m_ds[i][0] += real(.25);
    // Catmull-Clark version of DS
    m_ds[i][0] += real(.5); m_ds[i][1] += real(.125); m_ds[i][i-1] += real(.125);

    // BIQ rule
    for( j = 0; j < i; ++j ){
      m_bq[NW][i][j] = real(1)/(4*i)+(3*m_ds[i][j]+m_ds[i][(j+i-1)%i]+m_ds[i][(j+1)%i])/16;
      m_bq[SW][i][j] = (2*m_ds[i][j]+m_ds[i][(j+1)%i])/16;
      m_bq[SE][i][j] = m_ds[i][j]/16;
      m_bq[NE][i][j] = (2*m_ds[i][j]+m_ds[i][(j+i-1)%i])/16;
    }
  }
  m_init = true;
}

OneRing::OneRing( QTree *t, const Scheme sch )
  : m_c(t), m_s(sch)
{
  if(!m_init) BuildTables();
  
  m_t[N] = m_t[W] = m_t[S] = m_t[E] = 0;
  m_i[NW] = m_i[SW] = m_i[SE] = m_i[NE] = 0;
  
  // collect up the 1-ring of neighbors in all 4 quadrants; be prepared to
  // force subdivision if someone doesn't exist...
  // everyone has to live at the depth of t (t is the parent generation and
  // it is t's children we need to prepare for)
  unsigned int d = m_c->Depth();

  Perms Nperm, Wperm, Sperm, Eperm;
  // get the four neighbors at the proper depth
  if( ( m_t[N] = m_c->Neighbor( N, &Nperm ) ) && ( m_t[N]->Depth() < d ) ){
    m_t[N]->Subdiv(); m_t[N] = m_c->Neighbor( N, &Nperm ); 
    if(m_t[N]->Depth() != d) m_t[N] = 0;
  }
  assert( !m_t[N] || m_t[N]->Neighbor( Nperm[S] ) == m_c );

  if( ( m_t[W] = m_c->Neighbor( W, &Wperm ) ) && ( m_t[W]->Depth() < d ) ){
    m_t[W]->Subdiv(); m_t[W] = m_c->Neighbor( W, &Wperm ); 
    if(m_t[W]->Depth() != d) m_t[W] = 0;
  }
  assert( !m_t[W] || m_t[W]->Neighbor( Wperm[E] ) == m_c );

  if( ( m_t[S] = m_c->Neighbor( S, &Sperm ) ) && ( m_t[S]->Depth() < d ) ){
    m_t[S]->Subdiv(); m_t[S] = m_c->Neighbor( S, &Sperm ); 
    if(m_t[S]->Depth() != d) m_t[S] = 0;
  }  
  assert( !m_t[S] || m_t[S]->Neighbor( Sperm[N] ) == m_c );

  if( ( m_t[E] = m_c->Neighbor( E, &Eperm ) ) && ( m_t[E]->Depth() < d ) ){
    m_t[E]->Subdiv(); m_t[E] = m_c->Neighbor( E, &Eperm ); 
    if(m_t[E]->Depth() != d) m_t[E] = 0;
  }
  assert( !m_t[E] || m_t[E]->Neighbor( Eperm[W] ) == m_c );

  Perms perm; QTree *tv, *tvn; Direction prev;

  // NW quadrant
  if( m_t[N] && m_t[W] ){
    if( !t->Root() || t->Corner(NW)){
      // walk ccw from m_t[N] to m_t[W]
      bool orient = true;
      tv = m_t[N]; prev = Nperm[Next(N)];
      perm = Nperm;
      for( m_i[NW] = 0; tv != m_t[W]; prev = perm[orient?Next(prev):Prev(prev)] ){
	if( perm.Reflect() ) orient = !orient;
	tvn = tv->Neighbor( prev, &perm );
	if(!tvn){ m_i[NW] = -1; break; }
	else if( tvn->Depth() < d ){
	  tvn->Subdiv(); tvn = tv->Neighbor( prev, &perm ); assert( tvn->Depth() == d );
	}
	tv = tvn; assert( m_i[NW] < __MAXFACESIZE );
	if( tv != m_t[W] ){ m_q[NW][m_i[NW]] = tv; ++m_i[NW]; }
      }
    }else m_i[NW] = -1;
  }else m_i[NW] = -1;

  // SW
  if( m_t[W] && m_t[S] ){
    if( !t->Root() || t->Corner(SW) ){
      bool orient = true;
      tv = m_t[W]; prev = Wperm[Next(W)];
      perm = Wperm;
      for( m_i[SW] = 0; tv != m_t[S]; prev = perm[orient?Next(prev):Prev(prev)] ){
	if( perm.Reflect() ) orient = !orient;
	tvn = tv->Neighbor( prev, &perm );
	if(!tvn){ m_i[SW] = -1; break;}
	else if( tvn->Depth() < d ){
	  tvn->Subdiv(); tvn = tv->Neighbor( prev, &perm ); assert( tvn->Depth() == d );
	}
	tv = tvn; assert( m_i[SW] < __MAXFACESIZE );
	if( tv != m_t[S] ){ m_q[SW][m_i[SW]] = tv; ++m_i[SW]; }
      }
    }else m_i[SW] = -1;
  }else	m_i[SW] = -1;

  // SE
  if( m_t[S] && m_t[E] ){
    if( !t->Root() || t->Corner(SE) ){
      bool orient = true;
      tv = m_t[S]; prev = Sperm[Next(S)];
      perm = Sperm;
      for( m_i[SE] = 0; tv != m_t[E]; prev = perm[orient?Next(prev):Prev(prev)] ){
	if( perm.Reflect() ) orient = !orient;
	tvn = tv->Neighbor( prev, &perm );
	if(!tvn){ m_i[SE] = -1; break; }
	else if( tvn->Depth() < d ){
	  tvn->Subdiv(); tvn = tv->Neighbor( prev, &perm ); assert( tvn->Depth() == d );
	}
	tv = tvn; assert( m_i[SE] < __MAXFACESIZE );
	if( tv != m_t[E] ){ m_q[SE][m_i[SE]] = tv; ++m_i[SE]; }
      }
    }else m_i[SE] = -1;
  }else m_i[SE] = -1;

  // NE
  if( m_t[E] && m_t[N] ){
    if( !t->Root() || t->Corner(NE) ){
      bool orient = true;
      tv = m_t[E]; prev = Eperm[Next(E)];
      perm = Eperm;
      for( m_i[NE] = 0; tv != m_t[N]; prev = perm[orient?Next(prev):Prev(prev)] ){
	if( perm.Reflect() ) orient = !orient;
	tvn = tv->Neighbor( prev, &perm );
	if(!tvn){ m_i[NE] = -1; break;}
	else if ( tvn->Depth() < d ){
	  tvn->Subdiv(); tvn = tv->Neighbor( prev, &perm ); assert( tvn->Depth() == d );
	}
	tv = tvn; assert( m_i[NE] < __MAXFACESIZE );
	if( tv != m_t[N] ){ m_q[NE][m_i[NE]] = tv; ++m_i[NE]; }
      }
    }else m_i[NE] = -1;
  }else m_i[NE] = -1;
}

bool
OneRing::GetCorners( const QTreeName d, CVec *c[4] )
{
  c[NW] = c[SW] = c[SE] = c[NE] = 0;
  if( m_i[d] < 0 ) return false;

  QTree *n;
  Perms prm;
  switch( d ){
  case NW:
    c[NW] = m_i[NW] < 0 ? 0 : (assert(m_c->Corner(NW)), m_c->Corner(NW));

    // if one of the incident rings does not exist there will be no corner
    c[SW] = m_i[NW] < 0 || m_i[SW] < 0 ? 0 :
      // apparently they both exist; see whether this corner already exists
      // first check the other child if already there
      ( m_c->sw() ? (assert(m_c->sw()->Corner(NW)),m_c->sw()->Corner(NW)) :
	// the second clause will always be true IF executed and it will be
	// executed if n!=0 which it should be
	// maybe SE?
	( ( n = m_c->Neighbor( W, &prm ), assert(n), n->c(prm[SE]) ) ?
	  ( assert(n->c(prm[SE])->Corner(prm[NE])),n->c(prm[SE])->Corner(prm[NE]) ) :
	  // mabye NE
	  ( n->c(prm[NE]) ? (assert(n->c(prm[NE])->Corner(prm[SE])),n->c(prm[NE])->Corner(prm[SE])) : new CVec ) ) );
    
    c[SE] = m_i[NW] < 0 || m_i[SW] < 0 || m_i[SE] < 0 || m_i[NE] < 0 ? 0 :
      ( m_c->c(SW) ? (assert(m_c->c(SW)->Corner(NE)),m_c->c(SW)->Corner(NE)) :
	( m_c->c(SE) ? (assert(m_c->c(SE)->Corner(NW)),m_c->c(SE)->Corner(NW)) :
	  ( m_c->c(NE) ? (assert(m_c->c(NE)->Corner(SW)),m_c->c(NE)->Corner(SW)) : new CVec ) ) );
    
    c[NE] = m_i[NW] < 0 || m_i[NE] < 0 ? 0 :
      ( m_c->c(NE) ? (assert(m_c->c(NE)->Corner(NW)),m_c->c(NE)->Corner(NW)) :
	( ( n = m_c->Neighbor( N, &prm ), assert(n), n->c(prm[SW]) ) ?
	  ( assert(n->c(prm[SW])->Corner(prm[SE])),n->c(prm[SW])->Corner(prm[SE])) :
	  ( n->c(prm[SE]) ? (assert(n->c(prm[SE])->Corner(prm[SW])),n->c(prm[SE])->Corner(prm[SW])) : new CVec ) ) );
    
    break;
  case SW:
    c[NW] = m_i[NW] < 0 || m_i[SW] < 0 ? 0 :
    ( m_c->c(NW) ? (assert(m_c->c(NW)->Corner(SW)),m_c->c(NW)->Corner(SW)) :
      ( ( n = m_c->Neighbor( W, &prm ), assert(n), n->c(prm[SE]) ) ?
	( assert(n->c(prm[SE])->Corner(prm[NE])), n->c(prm[SE])->Corner(prm[NE])) :
	( n->c(prm[NE]) ? (assert(n->c(prm[NE])->Corner(prm[SE])), n->c(prm[NE])->Corner(prm[SE])) : new CVec ) ) );

    c[SW] = m_i[SW] < 0 ? 0 : (assert(m_c->Corner(SW)),m_c->Corner(SW));

    c[SE] = m_i[SW] < 0 || m_i[SE] < 0 ? 0 :
      ( m_c->c(SE) ? (assert(m_c->c(SE)->Corner(SW)),m_c->c(SE)->Corner(SW)) :
	( ( n = m_c->Neighbor( S, &prm ), assert(n), n->c(prm[NW]) ) ?
	  ( assert(n->c(prm[NW])->Corner(prm[NE])),n->c(prm[NW])->Corner(prm[NE])) :
	  ( n->c(prm[NE]) ? (assert(n->c(prm[NE])->Corner(prm[NW])),n->c(prm[NE])->Corner(prm[NW])) : new CVec ) ) );
    
    c[NE] = m_i[NW] < 0 || m_i[SW] < 0 || m_i[SE] < 0 || m_i[NE] < 0 ? 0 :
      ( m_c->c(NW) ? (assert(m_c->c(NW)->Corner(SE)),m_c->c(NW)->Corner(SE)) :
	( m_c->c(SE) ? (assert(m_c->c(SE)->Corner(NW)),m_c->c(SE)->Corner(NW)) :
	  ( m_c->c(NE) ? (assert(m_c->c(NE)->Corner(SW)),m_c->c(NE)->Corner(SW)) : new CVec ) ) );

    break;
  case SE:
    c[NW] = m_i[NW] < 0 || m_i[SW] < 0 || m_i[SE] < 0 || m_i[NE] < 0 ? 0 :
      ( m_c->c(NW) ? (assert(m_c->c(NW)->Corner(SE)),m_c->c(NW)->Corner(SE)) :
	( m_c->c(SW) ? (assert(m_c->c(SW)->Corner(NE)),m_c->c(SW)->Corner(NE)) :
	  ( m_c->c(NE) ? (assert(m_c->c(NE)->Corner(SW)),m_c->c(NE)->Corner(SW)) : new CVec ) ) );

    c[SW] = m_i[SW] < 0 || m_i[SE] < 0 ? 0 :
      ( m_c->c(SW) ? (assert(m_c->c(SW)->Corner(SE)),m_c->c(SW)->Corner(SE)) :
	( ( n = m_c->Neighbor( S, &prm ), assert(n), n->c(prm[NW]) ) ?
	  ( assert(n->c(prm[NW])->Corner(prm[NE])),n->c(prm[NW])->Corner(prm[NE])) :
	  ( n->c(prm[NE]) ? (assert(n->c(prm[NE])->Corner(prm[NW])),n->c(prm[NE])->Corner(prm[NW])) : new CVec ) ) );
    
    c[SE] = m_i[SE] < 0 ? 0 : (assert(m_c->Corner(SE)),m_c->Corner(SE));

    c[NE] = m_i[SE] < 0 || m_i[NE] < 0 ? 0 :
      ( m_c->c(NE) ? (assert(m_c->c(NE)->Corner(SE)),m_c->c(NE)->Corner(SE)) :
	( ( n = m_c->Neighbor( E, &prm ), assert(n), n->c(prm[NW]) ) ?
	  ( assert(n->c(prm[NW])->Corner(prm[SW])),n->c(prm[NW])->Corner(prm[SW])) :
	  ( n->c(prm[SW]) ? (assert(n->c(prm[SW])->Corner(prm[NW])),n->c(prm[SW])->Corner(prm[NW])) : new CVec ) ) );

    break;
  case NE:
    c[NW] = m_i[NW] < 0 || m_i[NE] < 0 ? 0 :
      ( m_c->c(NW) ? (assert(m_c->c(NW)->Corner(NE)),m_c->c(NW)->Corner(NE)) :
	( ( n = m_c->Neighbor( N, &prm ), assert(n), n->c(prm[SW]) ) ?
	  ( assert(n->c(prm[SW])->Corner(prm[SE])),n->c(prm[SW])->Corner(prm[SE])) :
	  ( n->c(prm[SE]) ? (assert(n->c(prm[SE])->Corner(prm[SW])),n->c(prm[SE])->Corner(prm[SW])) : new CVec ) ) );

    c[SW] = m_i[NW] < 0 || m_i[SW] < 0 || m_i[SE] < 0 || m_i[NE] < 0 ? 0 :
      ( m_c->c(NW) ? (assert(m_c->c(NW)->Corner(SE)),m_c->c(NW)->Corner(SE)) :
	( m_c->c(SW) ? (assert(m_c->c(SW)->Corner(NE)),m_c->c(SW)->Corner(NE)) :
	  ( m_c->c(SE) ? (assert(m_c->c(SE)->Corner(NW)),m_c->c(SE)->Corner(NW)) : new CVec ) ) );

    c[SE] = m_i[SE] < 0 || m_i[NE] < 0 ? 0 :
      ( m_c->c(SE) ? (assert(m_c->c(SE)->Corner(NE)),m_c->c(SE)->Corner(NE)) :
	( ( n = m_c->Neighbor( E, &prm ), assert(n), n->c(prm[NW]) ) ?
	  ( assert(n->c(prm[NW])->Corner(prm[SW])),n->c(prm[NW])->Corner(prm[SW])) :
	  ( n->c(prm[SW]) ? (assert(n->c(prm[SW])->Corner(prm[NW])),n->c(prm[SW])->Corner(prm[NW])) : new CVec ) ) );

    c[NE] = m_i[NE] < 0 ? 0 : (assert(m_c->Corner(NE)),m_c->Corner(NE));

    break;
  default:
    die();
  }

  return true;
}

// this is a VERY simple example of an estimator used for adaptive
// subdivision. It computes plane equations of the one-rings at the parent
// level and then measures the projections of the control pointsi n the
// one-ring against the normals, taking the min and max and returning their
// difference. One could play with other, perhaps much cheaper criteria for
// adpativity. Maybe something based on view extents, etc.... The sky is
// the limit here.
real
OneRing::Eps( QTree *t ) const
{
  // compute error estimate of closeness of current control mesh from
  // subdivision surface
  if( t->Leaf() ) return 0;

  real mx = 0, mn = 0, tmp;
  // compute plane estimates for the chains
  if( m_i[NW] >= 0 ){
    real t = 2*M_PI/(m_i[NW]+3);
    CVec p = m_c->loc() + m_t[N]->loc();
    CVec s = sin(0*t)*m_c->loc()+sin(1*t)*m_t[N]->loc();
    CVec c = cos(0*t)*m_c->loc()+cos(1*t)*m_t[N]->loc();
    int i;
    for( i = 0; i < m_i[NW]; ++i ){
      p += m_q[NW][i]->loc();
      s += sin((i+2)*t)*m_q[NW][i]->loc();
      c += cos((i+2)*t)*m_q[NW][i]->loc();
    }
    p += m_t[W]->loc();
    s += sin((i+2)*t)*m_t[W]->loc();
    c += cos((i+2)*t)*m_t[W]->loc();
    p /= m_i[NW]+3;
    
    CVec n = s.cross(c).dir();
    tmp = n.dot(m_c->loc()-p);
    mx = std::max( tmp, mx ); mn = std::min( tmp, mn );
    tmp = n.dot(m_t[N]->loc()-p);
    mx = std::max( tmp, mx ); mn = std::min( tmp, mn );
    for( i = 0; i < m_i[NW]; ++i ){
      tmp = n.dot(m_q[NW][i]->loc()-p);
      mx = std::max( tmp, mx ); mn = std::min( tmp, mn );
    }
    tmp = n.dot(m_t[W]->loc()-p);
    mx = std::max( tmp, mx ); mn = std::min( tmp, mn );
  }  
  if( m_i[SW] >= 0 ){
    real t = 2*M_PI/(m_i[SW]+3);
    CVec p = m_c->loc() + m_t[W]->loc();
    CVec s = sin(0*t)*m_c->loc()+sin(1*t)*m_t[W]->loc();
    CVec c = cos(0*t)*m_c->loc()+cos(1*t)*m_t[W]->loc();
    int i;
    for( i = 0; i < m_i[SW]; ++i ){
      p += m_q[SW][i]->loc();
      s += sin((i+2)*t)*m_q[SW][i]->loc();
      c += cos((i+2)*t)*m_q[SW][i]->loc();
    }
    p += m_t[S]->loc();
    s += sin((i+2)*t)*m_t[S]->loc();
    c += cos((i+2)*t)*m_t[S]->loc();
    p /= m_i[SW]+3;

    CVec n = s.cross(c).dir();
    tmp = n.dot(m_c->loc()-p);
    mx = std::max( tmp, mx ); mn = std::min( tmp, mn );
    tmp = n.dot(m_t[W]->loc()-p);
    mx = std::max( tmp, mx ); mn = std::min( tmp, mn );
    for( i = 0; i < m_i[SW]; ++i ){
      tmp = n.dot(m_q[SW][i]->loc()-p);
      mx = std::max( tmp, mx ); mn = std::min( tmp, mn );
    }
    tmp = n.dot(m_t[S]->loc()-p);
    mx = std::max( tmp, mx ); mn = std::min( tmp, mn );
  }  
  if( m_i[SE] >= 0 ){
    real t = 2*M_PI/(m_i[SE]+3);
    CVec p = m_c->loc() + m_t[S]->loc();
    CVec s = sin(0*t)*m_c->loc()+sin(1*t)*m_t[S]->loc();
    CVec c = cos(0*t)*m_c->loc()+cos(1*t)*m_t[S]->loc();
    int i;
    for( i = 0; i < m_i[SE]; ++i ){
      p += m_q[SE][i]->loc();
      s += sin((i+2)*t)*m_q[SE][i]->loc();
      c += cos((i+2)*t)*m_q[SE][i]->loc();
    }
    p += m_t[E]->loc();
    s += sin((i+2)*t)*m_t[E]->loc();
    c += cos((i+2)*t)*m_t[E]->loc();
    p /= m_i[SE]+3;

    CVec n = s.cross(c).dir();
    tmp = n.dot(m_c->loc()-p);
    mx = std::max( tmp, mx ); mn = std::min( tmp, mn );
    tmp = n.dot(m_t[S]->loc()-p);
    mx = std::max( tmp, mx ); mn = std::min( tmp, mn );
    for( i = 0; i < m_i[SE]; ++i ){
      tmp = n.dot(m_q[SE][i]->loc()-p);
      mx = std::max( tmp, mx ); mn = std::min( tmp, mn );
    }
    tmp = n.dot(m_t[E]->loc()-p);
    mx = std::max( tmp, mx ); mn = std::min( tmp, mn );
  }  
  if( m_i[NE] >= 0 ){
    real t = 2*M_PI/(m_i[NE]+3);
    CVec p = m_c->loc() + m_t[E]->loc();
    CVec s = sin(0*t)*m_c->loc()+sin(1*t)*m_t[E]->loc();
    CVec c = cos(0*t)*m_c->loc()+cos(1*t)*m_t[E]->loc();
    int i;
    for( i = 0; i < m_i[NE]; ++i ){
      p += m_q[NE][i]->loc();
      s += sin((i+2)*t)*m_q[NE][i]->loc();
      c += cos((i+2)*t)*m_q[NE][i]->loc();
    }
    p += m_t[N]->loc();
    s += sin((i+2)*t)*m_t[N]->loc();
    c += cos((i+2)*t)*m_t[N]->loc();
    p /= m_i[NE]+3;

    CVec n = s.cross(c).dir();
    tmp = n.dot(m_c->loc()-p);
    mx = std::max( tmp, mx ); mn = std::min( tmp, mn );
    tmp = n.dot(m_t[E]->loc()-p);
    mx = std::max( tmp, mx ); mn = std::min( tmp, mn );
    for( i = 0; i < m_i[NE]; ++i ){
      tmp = n.dot(m_q[NE][i]->loc()-p);
      mx = std::max( tmp, mx ); mn = std::min( tmp, mn );
    }
    tmp = n.dot(m_t[N]->loc()-p);
    mx = std::max( tmp, mx ); mn = std::min( tmp, mn );
  }
  return mx - mn;
}


CVec
OneRing::DooSabin( const QTreeName qtn ) const
{
  assert(c(qtn));
  if( m_i[qtn] == 1 ){
    return (real(9)/16)*m_c->loc()+
      (real(3)/16)*(m_t[First(qtn)]->loc()+m_t[Second(qtn)]->loc())+
      (real(1)/16)*m_q[qtn][0]->loc();
  }else{
    const int k = m_i[qtn] + 3; assert( k <= __MAXFACESIZE );
    const real *c = m_ds[k];
    CVec r = c[0]*m_c->loc()+c[1]*(m_t[First(qtn)]->loc()+m_t[Second(qtn)]->loc()); 
    for( int i = 2; i < k-1; ++i ) r += c[i]*m_q[qtn][i-2]->loc();
    return r;
  }
}

CVec
OneRing::MidEdge( const QTreeName qtn ) const
{
  assert(c(qtn));
  if( m_i[qtn] == 1 ){
    return (real(1)/2)*m_c->loc()+
      (real(1)/4)*(m_t[First(qtn)]->loc()+m_t[Second(qtn)]->loc());
  }else{
    const int k = m_i[qtn] + 3; assert( k <= __MAXFACESIZE );
    const real *c = m_me[k];
    CVec r = c[0]*m_c->loc()+c[1]*(m_t[First(qtn)]->loc()+m_t[Second(qtn)]->loc()); 
    for( int i = 2; i < k-1; ++i ) r += c[i]*m_q[qtn][i-2]->loc();
    return r;
  }
}

CVec
OneRing::Biquartic( const QTreeName qtn ) const
{
  assert(c(qtn));
  assert((m_i[NW]==1)+(m_i[SW]==1)+(m_i[SE]==1)+(m_i[NE]==1)>=3);
  if( ( m_i[NW] == 1 ) &&
      ( m_i[SW] == 1 ) &&
      ( m_i[SE] == 1 ) &&
      ( m_i[NE] == 1 ) ){
    return
      (real(100)/256)*m_c->loc()+
      (real(50)/256)*(m_t[First(qtn)]->loc()+m_t[Second(qtn)]->loc())+
      (real(10)/256)*(m_t[Oppos(First(qtn))]->loc()+m_t[Oppos(Second(qtn))]->loc())+
      (real(25)/256)*m_q[qtn][0]->loc()+
      (real(5)/256)*(m_q[Next(qtn)][0]->loc()+m_q[Prev(qtn)][0]->loc())+
      (real(1)/256)*m_q[Next(Next(qtn))][0]->loc();
  }else{
    // rotate extraordinary face into NW
    const QTreeName qn =  m_i[NW] != 1 ? NW : ( m_i[SW] != 1 ? SW : ( m_i[SE] != 1 ? SE : NE ) );
    const int k = m_i[qn] + 3; assert( k <= __MAXFACESIZE );
    const QTreeName st = Dist(qtn,qn);
    const real *c = m_bq[st][k];
    CVec r =
      (m_fbq[st][0]+c[0])*m_c->loc()+
      (m_fbq[st][1]+c[1])*m_t[First(qn)]->loc()+
      (m_fbq[st][2]+c[k-1])*m_t[Second(qn)]->loc()+
      m_fbq[st][3]*m_q[Prev(qn)][0]->loc()+
      m_fbq[st][4]*m_q[Next(qn)][0]->loc()+
      m_fbq[st][5]*m_t[Oppos(Second(qn))]->loc()+
      m_fbq[st][6]*m_t[Oppos(First(qn))]->loc()+
      m_fbq[st][7]*m_q[Next(Next(qn))][0]->loc();
    for( int i = 2; i < k-1; ++i ) r += c[i]*m_q[qn][i-2]->loc();
    return r;
  }
}
