// -*- Mode: c++ -*-
// $Id: QTree.cpp,v 1.9 2000/04/28 22:14:47 ps Exp $
// $Source: /ufs/multires/projects//src/CourseQuad/QTree.cpp,v $

/**
 * Copyright 1995-2000, Peter Schr\"oder, California Institute of
 * Technology (Caltech) 
 *
 * April 26, 2000
 * All Rights Reserved
 * 
 * The use of this software is permitted for non-commercial, educational,
 * and research use only. The software and/or related materials are
 * provided "as-is" without warranty of any kind including any warranties
 * of performance or merchantability or fitness for a particular use or
 * purpose (as set forth in UCC 23212-2313) or for any purpose
 * whatsoever, for the licensed product, however used. In no event shall
 * Caltech be liable for any damages and/or costs, including but not
 * limited to incidental or consequential damages of any kind, including
 * economic damage or injury to property and lost profits, regardless of
 * whether Caltech shall be advised, have reason to know, or in fact
 * shall know of the possibility. User bears all risk relating to quality
 * and performance of the software and/or related materials.
 * 
 * Any use other than non-commercial, educational, or research, or any
 * redistribution in original or modified form requires prior written
 * authorization from the copyright holder.
 */

#include <algorithm>
#include "QTree.h"
#include "Cp.h"
#include "OneRing.h"
#include "Perms.h"

QTree::QTree( Cp *cp,
	      QTree *parent,
	      CVec *corners[4],
	      const CVec& location,
	      const QTreeName name,
	      const unsigned int depth )
  : m_cp( cp ),
    m_p( parent ),
    m_l( location ),
    m_eps( 0 ),
    m_depth( depth ),
    m_n( name )
{
  assert( m_cp ); assert( m_depth < __MAXDEPTH );
  m_c[NE] = m_c[NW] = m_c[SW] = m_c[SE] = 0;
  if( corners ){
    m_corner[NE] = corners[NE];
    m_corner[NW] = corners[NW];
    m_corner[SW] = corners[SW];
    m_corner[SE] = corners[SE];
  }else m_corner[NE] = m_corner[NW] = m_corner[SW] = m_corner[SE] = 0;
}

QTree::~QTree( void )
{
  // release edge midpoints on the way down the recursion
  Perms np;
  QTree *n;
  // NW
  if( m_c[NW] ){
    // center
    delete m_c[NW]->m_corner[SE]; m_c[NW]->m_corner[SE] = 0;
    if( m_c[SW] ) m_c[SW]->m_corner[NE] = 0;
    if( m_c[SE] ) m_c[SE]->m_corner[NW] = 0;
    if( m_c[NE] ) m_c[NE]->m_corner[SW] = 0;
    
    // northern edge
    if( !( n = Neighbor( N, &np ) ) ||
	( n->Depth() != Depth() ) ||
	( !n->m_c[np[SW]] && !n->m_c[np[SE]] ) ){
      // no neighbor, not of the same depth, neighbor of same depth with no
      // children sharing this edge
      delete m_c[NW]->m_corner[NE]; m_c[NW]->m_corner[NE] = 0;
      if( m_c[NE] ) m_c[NE]->m_corner[NW] = 0;
    }else{
      m_c[NW]->m_corner[NE] = 0;
      if( m_c[NE] ) m_c[NE]->m_corner[NW] = 0;
    }
    // western edge
    if( !( n = Neighbor( W, &np ) ) ||
	( n->Depth() != Depth() ) ||
	( !n->m_c[np[NE]] && !n->m_c[np[SE]] ) ){
      delete m_c[NW]->m_corner[SW]; m_c[NW]->m_corner[SW] = 0;
      if( m_c[SW] ) m_c[SW]->m_corner[NW] = 0;
    }else{
      m_c[NW]->m_corner[SW] = 0;
      if( m_c[SW] ) m_c[SW]->m_corner[NW] = 0;
    }
  }

  // SW
  if( m_c[SW] ){
    // center
    delete m_c[SW]->m_corner[NE]; m_c[SW]->m_corner[NE] = 0;
    if( m_c[SE] ) m_c[SE]->m_corner[NW] = 0;
    if( m_c[NE] ) m_c[NE]->m_corner[SW] = 0;
    if( m_c[NW] ) m_c[NW]->m_corner[SE] = 0;
    
    // western edge
    if( !( n = Neighbor( W, &np ) ) ||
	( n->Depth() != Depth() ) ||
	( !n->m_c[np[NE]] && !n->m_c[np[SE]] ) ){
      delete m_c[SW]->m_corner[NW]; m_c[SW]->m_corner[NW] = 0;
      if( m_c[NW] ) m_c[NW]->m_corner[SW] = 0;
    }else{
      m_c[SW]->m_corner[NW] = 0;
      if( m_c[NW] ) m_c[NW]->m_corner[SW] = 0;
    }
    // southern edge
    if( !( n = Neighbor( S, &np ) ) ||
	( n->Depth() != Depth() ) ||
	( !n->m_c[np[NW]] && !n->m_c[np[NE]] ) ){
      delete m_c[SW]->m_corner[SE]; m_c[SW]->m_corner[SE] = 0;
      if( m_c[SE] ) m_c[SE]->m_corner[SW] = 0;
    }else{
      m_c[SW]->m_corner[SE] = 0;
      if( m_c[SE] ) m_c[SE]->m_corner[SW] = 0;
    }
  }
    
  // SE
  if( m_c[SE] ){
    // center
    delete m_c[SE]->m_corner[NW]; m_c[SE]->m_corner[NW] = 0;
    if( m_c[NE] ) m_c[NE]->m_corner[SW] = 0;
    if( m_c[NW] ) m_c[NW]->m_corner[SE] = 0;
    if( m_c[SW] ) m_c[SW]->m_corner[NE] = 0;
    
    // southern edge
    if( !( n = Neighbor( S, &np ) ) ||
	( n->Depth() != Depth() ) ||
	( !n->m_c[np[NW]] && !n->m_c[np[NE]] ) ){
      delete m_c[SE]->m_corner[SW]; m_c[SE]->m_corner[SW] = 0;
      if( m_c[SW] ) m_c[SW]->m_corner[SE] = 0;
    }else{
      m_c[SE]->m_corner[SW] = 0;
      if( m_c[SW] ) m_c[SW]->m_corner[SE] = 0;
    }
    // eastern edge
    if( !( n = Neighbor( E, &np ) ) ||
	( n->Depth() != Depth() ) ||
	( !n->m_c[np[NW]] && !n->m_c[np[SW]] ) ){
      delete m_c[SE]->m_corner[NE]; m_c[SE]->m_corner[NE] = 0;
      if( m_c[NE] ) m_c[NE]->m_corner[SE] = 0;
    }else{
      m_c[SE]->m_corner[NE] = 0;
      if( m_c[NE] ) m_c[NE]->m_corner[SE] = 0;
    }
  }

  // NE
  if( m_c[NE] ){
    // center
    delete m_c[NE]->m_corner[SW]; m_c[NE]->m_corner[SW] = 0;
    if( m_c[NW] ) m_c[NW]->m_corner[SE] = 0;
    if( m_c[SW] ) m_c[SW]->m_corner[NE] = 0;
    if( m_c[SE] ) m_c[SE]->m_corner[NW] = 0;
    
    // eastern edge
    if( !( n = Neighbor( E, &np ) ) ||
	( n->Depth() != Depth() ) ||
	( !n->m_c[np[NW]] && !n->m_c[np[SW]] ) ){
      delete m_c[NE]->m_corner[SE]; m_c[NE]->m_corner[SE] = 0;
      if( m_c[SE] ) m_c[SE]->m_corner[NE] = 0;
    }else{
      m_c[NE]->m_corner[SE] = 0;
      if( m_c[SE] ) m_c[SE]->m_corner[NE] = 0;
    }
    // northern edge
    if( !( n = Neighbor( N, &np ) ) ||
	( n->Depth() != Depth() ) ||
	( !n->m_c[np[SW]] && !n->m_c[np[SE]] ) ){
      delete m_c[NE]->m_corner[NW]; m_c[NE]->m_corner[NW] = 0;
      if( m_c[NW] ) m_c[NW]->m_corner[NE] = 0;
    }else{
      m_c[NE]->m_corner[NW] = 0;
      if( m_c[NW] ) m_c[NW]->m_corner[NE] = 0;
    }
  }
  
  delete m_c[NW]; m_c[NW] = 0;
  delete m_c[SW]; m_c[SW] = 0;
  delete m_c[SE]; m_c[SE] = 0;
  delete m_c[NE]; m_c[NE] = 0;

  // on the way up: notify parent 
  if( m_p ){
    if      ( m_p->m_c[NW] == this )       m_p->m_c[NW] = 0;
    else if ( m_p->m_c[SW] == this )       m_p->m_c[SW] = 0;
    else if ( m_p->m_c[SE] == this )       m_p->m_c[SE] = 0;
    else ( assert( m_p->m_c[NE] == this ), m_p->m_c[NE] = 0 );
  }else{
    // I am root and I should do away with my 4 corners
    m_corner[NW] = m_corner[SW] = m_corner[SE] = m_corner[NE] = 0;
  }
}

bool
QTree::Subdiv( const real eps )
{
  // only if leaf; always when root, otherwise parent must fail eps test
  if( Leaf() && ( Root() || ( m_p->m_eps > eps ) ) ){
    // get the complete 1-ring of neighbors, possibly creating them
    // recursively if needed
    OneRing r( this, CP()->scheme() );
    
    // ok; at least some children get corners
    CVec *c[4];

    m_c[NW] = r.GetCorners( NW, c ) ? new QTree( m_cp, this, c, r.Subdiv(NW), NW, m_depth + 1 ) : 0;
    // assert successful creation of child if in fact warranted
    assert( !r.c(NW) || m_c[NW] );
    // verify the corners
    assert( ( !!c[NW] == !!r.c(NW) ) &&
	    ( !!c[SW] == ( r.c(NW) && r.c(SW) ) ) &&
	    ( !!c[SE] == ( r.c(NW) && r.c(SW) && r.c(SE) && r.c(NE) ) ) &&
	    ( !!c[NE] == ( r.c(NW) && r.c(NE) ) ) );

    m_c[SW] = r.GetCorners( SW, c ) ? new QTree( m_cp, this, c, r.Subdiv(SW), SW, m_depth + 1 ) : 0;
    assert( !r.c(SW) || m_c[SW] );
    assert( ( !!c[NW] == ( r.c(NW) && r.c(SW) ) ) &&
	    ( !!c[SW] == !!r.c(SW) ) &&
	    ( !!c[SE] == ( r.c(SW) && r.c(SE) ) ) &&
	    ( !!c[NE] == ( r.c(NW) && r.c(SW) && r.c(SE) && r.c(NE) ) ) );

    m_c[SE] = r.GetCorners( SE, c ) ? new QTree( m_cp, this, c, r.Subdiv(SE), SE, m_depth + 1 ) : 0;
    assert( !r.c(SE) || m_c[SE] );
    assert( ( !!c[NW] == ( r.c(NW) && r.c(SW) && r.c(SE) && r.c(NE) ) ) &&
	    ( !!c[SW] == ( r.c(SW) && r.c(SE) ) ) &&
	    ( !!c[SE] == !!r.c(SE) ) &&
	    ( !!c[NE] == ( r.c(SE) && r.c(NE) ) ) );

    m_c[NE] = r.GetCorners( NE, c ) ? new QTree( m_cp, this, c, r.Subdiv(NE), NE, m_depth + 1 ) : 0;
    assert( !r.c(NE) || m_c[NE] );
    assert( ( !!c[NW] == ( r.c(NW) && r.c(NE) ) ) &&
	    ( !!c[SW] == ( r.c(NW) && r.c(SW) && r.c(SE) && r.c(NE) ) ) &&
	    ( !!c[SE] == ( r.c(SE) && r.c(NE) ) ) &&
	    ( !!c[NE] == !!r.c(NE) ) );

    m_eps = r.Eps( this );
  }

  // successful subdivision
  return true;
}

void
QTree::Expand( const unsigned int depth, const real eps )
{
  if( depth > 0 ){
    Subdiv( eps ); 
    if( m_c[NW] ) m_c[NW]->Expand( depth - 1, eps );
    if( m_c[SW] ) m_c[SW]->Expand( depth - 1, eps );
    if( m_c[SE] ) m_c[SE]->Expand( depth - 1, eps );
    if( m_c[NE] ) m_c[NE]->Expand( depth - 1, eps );
  }
}

// given a direction we will flip the name of that direction (or its
// opposite) in the quadrant name. E.g., for d==N NW->SW, NE->SE, SW->NW,
// SE->NE
QTreeName QTree::m_reflect[4/*Direction*/][4/*QTreeName*/] =
{ //NW  SW  SE  NE
  { SW, NW, NE, SE }, // N
  { NE, SE, SW, NW }, // W
  { SW, NW, NE, SE }, // S
  { NE, SE, SW, NW }  // E
};

// is the neighbor in the direction d from qtn in this parent?
bool QTree::m_opposite[4/*Direction*/][4/*QTreeName*/] =
{ //  NW    SW    SE    NE
  { false, true, true, false }, // N
  { false, false, true, true }, // W
  { true, false, false, true }, // S
  { true, true, false, false }  // E
};

// once we know that a neighbor in a given direction from our present
// location can be found within this level of the present quadtree,
// retrieve its name
const QTreeName XX = QTreeName(-1);
QTreeName QTree::m_crossover[4/*Direction*/][4/*QTreeName*/] =
{ //NW  SW  SE  NE
  { XX, NW, NE, XX }, // N
  { XX, XX, SW, NW }, // W
  { SW, XX, XX, SE }, // S
  { NE, SE, XX, XX }  // E
};

// find the neighbor in a given direction which is no smaller than `this'
// basically we ascend the tree to the least common ancestor at which we
// can cross over. Then we cross over and descend again (if possible) to
// the original depth, this time in a reflected manner. 

QTree*
QTree::Neighbor( const Direction d, Perms * perm ) const
{
  // starting with ourselves
  QTree *qt = (QTree *)this;
  // remember the names of children ascended
  unsigned long int names = 0;
  // we don't have more room in `names'
  assert( m_depth <= (((sizeof names)*8 ) >> 2));
  // count the steps taken
  unsigned int i = 0;
  const unsigned int UP = 0, DOWN = 1;
  // record the direction
  unsigned int dir = UP;

  // set default to identity
  Perms nPerm = Perms::Ident();

  // ascent
  while( dir == UP ){
    // name of this one
    QTreeName qtn = qt->n();
    if( qt->Root() ){
      // at the root we can always cross over
      // in preparation for the descent we have to change all the names to
      // the permutation which lives on the other side
      if( !qt->m_cp->Permute( names, i, d, &nPerm ) ){
	// no such neighbor 
	return 0;
      }
      // now do the jump to the other guy
      qt = qt->m_cp->Neighbor( d )->Root();
      dir = DOWN;
    }else if( m_opposite[d][qtn] ){
      // ok we can jump over
      qt = qt->m_p->m_c[m_crossover[d][qtn]];
      dir = DOWN;
    }else{
      // can't jump yet; need to ascend further
      // remember the quadtree name for the descent and count steps
      names = ( names << 2 ) | qtn; i++;
      // move up the qtree we are not at the root, so we can ascend further
      qt = qt->m_p; assert( qt );
    }
  }

  // descent
  while( qt && i-- ){
    // on the way down all the names are reflected about the direction we
    // crossed over
    QTreeName qtn = m_reflect[nPerm[d]][QTreeName( names & 0x3 )];
    // get the appropriate child or if you can't, remember the last one
    if( !qt->m_c[qtn] ) break; else qt = qt->m_c[qtn];
    // toss the name
    names >>= 2;
  }

  // return the value of the permutation of the neighbor
  if( perm ) *perm = nPerm;

  // if we didn't find a child at our level, i will tell us how many levels
  // we are off because of the `break' statement above.
  // this information is not currently used, however

  return qt;
}

int
QTree::QTreeDepth( void ) const
{
  // stupid but effective...
  if( Leaf() )  return m_depth;
  else{
    return std::max( std::max( m_c[NW] ? m_c[NW]->QTreeDepth() : m_depth,
			       m_c[SW] ? m_c[SW]->QTreeDepth() : m_depth ),
		     std::max( m_c[SE] ? m_c[SE]->QTreeDepth() : m_depth,
			       m_c[NE] ? m_c[NE]->QTreeDepth() : m_depth ) );
  }
}
